import re

from . import categorytype


class CharacterCategory(object):
    class Range(object):
        def __init__(self):
            self.low = 0
            self.high = 0
            self.categories = set()

        def contains(self, cp):
            if cp >= self.low and cp <= self.high:
                return True
            return False

        def containing_length(self, text):
            for i in range(len(text)):
                c = ord(text[i])
                if c < self.low or c > self.high:
                    return i
            return len(text)

    def __init__(self):
        self.range_list = []

    def get_category_types(self, code_point):
        for range_ in self.range_list:
            if range_.contains(code_point):
                return range_.categories
        return set(categorytype.CategoryType.DEFAULT)

    def read_character_definition(self, char_def=None):
        """
        :param char_def: path
        """

        if char_def is not None:
            f = open(char_def, 'r', encoding="utf-8")
        else:
            f = open("char.def", 'r', encoding="utf-8")

        for i, line in enumerate(f.readlines()):
            line = line.rstrip()
            if re.fullmatch(r"\s*", line) or re.match("#", line):
                continue
            cols = re.split(r"\s+", line)
            if len(cols) < 2:
                f.close()
                raise AttributeError("invalid format at line {}".format(i))
            if not re.match("0x", cols[0]):
                continue
            range_ = self.Range()
            r = re.split("\\.\\.", cols[0])
            range_.low = range_.high = int(r[0], 16)
            if len(r) > 1:
                range_.high = int(r[1], 16)
            if range_.low > range_.high:
                f.close()
                raise AttributeError("invalid range at line {}".format(i))
            for j in range(1, len(cols)):
                if re.match("#", cols[j]) or cols[j] is '':
                    break
                type_ = categorytype.CategoryType.get(cols[j])
                if type_ is None:
                    f.close()
                    raise AttributeError("{} is invalid type at line {}".format(cols[j], i))
                range_.categories.add(type_)
            self.range_list.append(range_)
        default_range = self.Range()
        default_range.low = 0
        default_range.high = float('inf')
        default_range.categories.add(categorytype.CategoryType.DEFAULT)
        self.range_list.reverse()
        self.range_list.append(default_range)

        f.close()



from .dictionarylib.wordinfo import WordInfo


class LatticeNode:

    __NULL_SURFACE = '(null)'

    begin = 0
    end = 0
    total_cost = 0
    word_id = 0
    _is_oov = False
    best_previous_node = None
    is_connected_to_bos = None
    extra_word_info = None

    def __init__(self, lexicon=None, left_id=None, right_id=None, cost=None, word_id=None):

        self.undefined_word_info =\
            WordInfo(self.__NULL_SURFACE, 0, -1, self.__NULL_SURFACE, -1,
                     self.__NULL_SURFACE, self.__NULL_SURFACE, [], [], [])
        self.is_defined = True
        if lexicon is left_id is right_id is cost is word_id is None:
            self.is_defined = False
            return
        self.lexicon = lexicon
        self.left_id = left_id
        self.right_id = right_id
        self.cost = cost
        self.word_id = word_id

    def set_parameter(self, left_id: int, right_id: int, cost: int) -> None:
        self.left_id = left_id
        self.right_id = right_id
        self.cost = cost

    def get_begin(self) -> int:
        return self.begin

    def get_end(self) -> int:
        return self.end

    def set_range(self, begin: int, end: int) -> None:
        self.begin = begin
        self.end = end

    def is_oov(self):
        return self._is_oov

    def set_oov(self):
        self._is_oov = True

    def get_word_info(self) -> WordInfo:
        if not self.is_defined:
            return self.undefined_word_info
        if self.extra_word_info:
            return self.extra_word_info
        return self.lexicon.get_word_info(self.word_id)

    def set_word_info(self, word_info: WordInfo) -> None:
        self.extra_word_info = word_info
        self.is_defined = True

    def get_path_cost(self) -> int:
        return self.cost

    def get_word_id(self) -> int:
        return self.word_id

    def get_dictionary_id(self) -> int:
        if not self.is_defined or self.extra_word_info:
            return -1
        return self.lexicon.get_dictionary_id(self.word_id)  # self.word_id >> 28

    def __str__(self):
        surface = ""
        if self.word_id < 0 and self.extra_word_info is None:
            surface = "(None)"
        else:
            surface = self.get_word_info().surface

        return "{} {} {}({}) {} {} {}".format(
            self.get_begin(), self.get_end(), surface, self.word_id, self.left_id, self.right_id, self.cost
        )



from .join_katakana_oov_plugin import JoinKatakanaOovPlugin
from .join_numeric_plugin import JoinNumericPlugin
from .path_rewrite_plugin import PathRewritePlugin
from .utils import get_path_rewrite_plugins



from enum import Enum


class CategoryType(Enum):
    DEFAULT = 1
    SPACE = 1 << 1
    KANJI = 1 << 2
    SYMBOL = 1 << 3
    NUMERIC = 1 << 4
    ALPHA = 1 << 5
    HIRAGANA = 1 << 6
    KATAKANA = 1 << 7
    KANJINUMERIC = 1 << 8
    GREEK = 1 << 9
    CYRILLIC = 1 << 10
    USER1 = 1 << 11
    USER2 = 1 << 12
    USER3 = 1 << 13
    USER4 = 1 << 14
    NOOOVBOW = 1 << 15

    def get_id(self):
        return self.id

    def get_type(self, id_):
        for type_ in CategoryType.values():
            if type_.get_id() is id_:
                return type_
        return None

    @staticmethod
    def get(str_):
        try:
            return CategoryType[str_]
        except KeyError:
            return None



import mmap

from . import SYSTEM_DICT_VERSION, USER_DICT_VERSION_1, USER_DICT_VERSION_2
from .dictionaryheader import DictionaryHeader
from .doublearraylexicon import DoubleArrayLexicon
from .grammar import Grammar


class BinaryDictionary(object):

    def __init__(self, bytes_: mmap.mmap, grammar: Grammar, header: DictionaryHeader, lexicon: DoubleArrayLexicon):
        self._bytes = bytes_
        self._grammar = grammar
        self._header = header
        self._lexicon = lexicon

    @staticmethod
    def _read_dictionary(filename, access=mmap.ACCESS_READ):
        with open(filename, 'r+b') as system_dic:
            bytes_ = mmap.mmap(system_dic.fileno(), 0, access=access)
        offset = 0
        header = DictionaryHeader.from_bytes(bytes_, offset)
        offset += header.storage_size()
        if header.version not in [SYSTEM_DICT_VERSION, USER_DICT_VERSION_1, USER_DICT_VERSION_2]:
            raise Exception('invalid dictionary version')
        grammar = None
        if header.version != USER_DICT_VERSION_1:
            grammar = Grammar(bytes_, offset)
            offset += grammar.get_storage_size()

        lexicon = DoubleArrayLexicon(bytes_, offset)
        return bytes_, grammar, header, lexicon

    @classmethod
    def from_system_dictionary(cls, filename):
        args = cls._read_dictionary(filename)
        version = args[2].version
        if version != SYSTEM_DICT_VERSION:
            raise IOError('invalid system dictionary')
        return cls(*args)

    @classmethod
    def from_user_dictionary(cls, filename):
        args = cls._read_dictionary(filename, mmap.ACCESS_COPY)
        version = args[2].version
        if version not in [USER_DICT_VERSION_1, USER_DICT_VERSION_2]:
            raise IOError('invalid user dictionary')
        return cls(*args)

    def close(self):
        self._bytes.close()

    @property
    def grammar(self) -> Grammar:
        return self._grammar

    @property
    def header(self) -> DictionaryHeader:
        return self._header

    @property
    def lexicon(self) -> DoubleArrayLexicon:
        return self._lexicon



import os
from collections import defaultdict

from sudachipy import config
from sudachipy.dictionarylib import categorytype
from sudachipy.dictionarylib import wordinfo

from .oov_provider_plugin import OovProviderPlugin


class MeCabOovPlugin(OovProviderPlugin):
    class CategoryInfo:
        def __init__(self):
            self.type_ = None
            self.is_invoke = None
            self.is_group = None
            self.length = None

    class OOV:
        def __init__(self):
            self.left_id = None
            self.right_id = None
            self.cost = None
            self.pos_id = None

    def __init__(self, json_obj=None):
        self.__chardef_filename = json_obj['charDef']
        self.__unkdef_filename = json_obj['unkDef']
        self.categories = {}
        self.oov_list = defaultdict(list)

    def set_up(self, grammar):
        char_def = os.path.join(config.settings.resource_dir, self.__chardef_filename)
        if not char_def:
            raise AttributeError("charDef is not defined")
        self.read_character_property(char_def)

        unk_def = os.path.join(config.settings.resource_dir, self.__unkdef_filename)
        if not unk_def:
            raise AttributeError("unkDef is not defined")
        self.read_oov(unk_def, grammar)

    def provide_oov(self, input_text, offset, has_other_words):
        nodes = []
        length = input_text.get_char_category_continuous_length(offset)
        if length > 0:
            for type_ in input_text.get_char_category_types(offset):
                if type_ not in self.categories:
                    continue
                cinfo = self.categories[type_]
                llength = length
                if cinfo.type_ not in self.oov_list:
                    continue
                oovs = self.oov_list[cinfo.type_]
                if cinfo.is_group and (cinfo.is_invoke or not has_other_words):
                    s = input_text.get_substring(offset, offset + length)
                    for oov in oovs:
                        nodes.append(self.get_oov_node(s, oov, length))
                        llength = -1
                if cinfo.is_invoke or not has_other_words:
                    for i in range(1, cinfo.length + 1):
                        sublength = input_text.get_code_points_offset_length(offset, i)
                        if sublength > llength:
                            break
                        s = input_text.get_substring(offset, offset + sublength)
                        for oov in oovs:
                            nodes.append(self.get_oov_node(s, oov, sublength))
        return nodes

    def get_oov_node(self, text, oov, length):
        node = self.create_node()
        node.set_parameter(oov.left_id, oov.right_id, oov.cost)
        info = wordinfo.WordInfo(surface=text, head_word_length=length, pos_id=oov.pos_id, normalized_form=text,
                                 dictionary_form_word_id=-1, dictionary_form=text, reading_form="",
                                 a_unit_split=[], b_unit_split=[], word_structure=[])
        node.set_word_info(info)
        return node

    def read_character_property(self, char_def):
        with open(char_def, "r", encoding="utf-8") as f:
            for i, line in enumerate(f, start=1):
                line = line.strip()
                if not line or line.startswith("#"):
                    continue
                cols = line.split()
                if len(cols) < 2:
                    raise RuntimeError("invalid format at line {}".format(i))
                if cols[0].startswith("0x"):
                    continue
                try:
                    type_ = getattr(categorytype.CategoryType, cols[0])
                except AttributeError:
                    raise RuntimeError("`{}` is invalid type at line {}".format(cols[0], i))
                if type_ in self.categories:
                    raise RuntimeError("`{}` is already defined at line {}".format(cols[0], i))

                info = self.CategoryInfo()
                info.type_ = type_
                info.is_invoke = (cols[1] != "0")
                info.is_group = (cols[2] != "0")
                info.length = int(cols[3])
                self.categories[type_] = info

    def read_oov(self, unk_def, grammar):
        with open(unk_def, "r", encoding="utf-8") as f:
            for i, line in enumerate(f, start=1):
                line = line.strip()
                if not line:
                    continue
                cols = line.split(",")
                if len(cols) < 10:
                    raise RuntimeError("invalid format at line {}".format(i))
                try:
                    type_ = getattr(categorytype.CategoryType, cols[0])
                except AttributeError:
                    raise RuntimeError("`{}` is invalid type at line {}".format(cols[0], i))
                if type_ not in self.categories:
                    raise RuntimeError("`{}` is undefined at line {}".format(cols[0], i))

                oov = self.OOV()
                oov.left_id = int(cols[1])
                oov.right_id = int(cols[2])
                oov.cost = int(cols[3])
                pos = cols[4:10]
                oov.pos_id = grammar.get_part_of_speech_id(pos)
                self.oov_list[type_].append(oov)



from abc import ABC, abstractmethod

from sudachipy.dictionarylib.grammar import Grammar


class EditConnectionCostPlugin(ABC):

    @abstractmethod
    def set_up(self, grammar: Grammar) -> None:
        raise NotImplementedError

    @abstractmethod
    def edit(self, grammar: Grammar) -> None:
        raise NotImplementedError

    @staticmethod
    def inhibit_connection(grammar: Grammar, left_id: int, right_id: int) -> None:
        grammar.set_connect_cost(left_id, right_id, Grammar.INHIBITED_CONNECTION)



import csv
import re
from logging import DEBUG, StreamHandler, getLogger

from sortedcontainers import SortedDict

from sudachipy.dartsclone.doublearray import DoubleArray
from sudachipy.dictionarylib.jtypedbytebuffer import JTypedByteBuffer
from sudachipy.dictionarylib.wordinfo import WordInfo


class DictionaryBuilder(object):

    __BYTE_MAX_VALUE = 127
    __MAX_LENGTH = 255
    __COLS_NUM = 18
    __BUFFER_SIZE = 1024 * 1024
    __PATTERN_UNICODE_LITERAL = re.compile(r"\\u([0-9a-fA-F]{4}|{[0-9a-fA-F]+})")
    __ARRAY_MAX_LENGTH = __BYTE_MAX_VALUE  # max value of byte in Java
    __STRING_MAX_LENGTH = 32767  # max value of short in Java
    is_user_dictionary = False

    class WordEntry:
        headword = None
        parameters = None
        wordinfo = None
        aunit_split_string = None
        bunit_split_string = None
        cunit_split_string = None

    class PosTable(object):

        def __init__(self):
            self.table = []

        def get_id(self, str_):
            id_ = self.table.index(str_) if str_ in self.table else -1
            if id_ < 0:
                id_ = len(self.table)
                self.table.append(str_)
            return id_

        def get_list(self):
            return self.table

    @staticmethod
    def __default_logger():
        handler = StreamHandler()
        handler.terminator = ""
        handler.setLevel(DEBUG)
        logger = getLogger(__name__)
        logger.setLevel(DEBUG)
        logger.addHandler(handler)
        logger.propagate = False
        return logger

    def __init__(self, *, logger=None):
        self.byte_buffer = JTypedByteBuffer()
        self.trie_keys = SortedDict()
        self.entries = []
        self.is_dictionary = False
        self.pos_table = self.PosTable()
        self.logger = logger or self.__default_logger()

    def build(self, lexicon_paths, matrix_input_stream, out_stream):
        self.logger.info('reading the source file...')
        for path in lexicon_paths:
            with open(path) as rf:
                self.build_lexicon(rf)
        self.logger.info('{} words\n'.format(len(self.entries)))

        self.write_grammar(matrix_input_stream, out_stream)
        self.write_lexicon(out_stream)

    def build_lexicon(self, lexicon_input_stream):
        line_no = -1
        try:
            for i, row in enumerate(csv.reader(lexicon_input_stream)):
                line_no = -1
                entry = self.parse_line(row)
                if entry.headword:
                    self.add_to_trie(entry.headword, len(self.entries))
                self.entries.append(entry)
        except Exception as e:
            if line_no > 0:
                self.logger.error(
                    '{} at line {} in {}\n'.format(e.args[0], line_no, lexicon_input_stream.name))
            raise e

    def parse_line(self, cols):
        if len(cols) != self.__COLS_NUM:
            raise ValueError('invalid format')
        cols = [self.decode(col) for col in cols]
        if not self.__is_length_valid(cols):
            raise ValueError('string is too long')
        if not cols[0]:
            raise ValueError('headword is empty')

        entry = self.WordEntry()
        # head word for trie
        if cols[1] != '-1':
            entry.headword = cols[0]
        # left-id, right-id, connect_cost
        entry.parameters = [int(cols[i]) for i in [1, 2, 3]]
        # part of speech
        pos_id = self.get_posid(cols[5:11])
        if pos_id < 0:
            raise ValueError('invalid part of speech')

        entry.aunit_split_string = cols[15]
        entry.bunit_split_string = cols[16]
        entry.cunit_split_string = cols[17]
        self.check_splitinfo_format(entry.aunit_split_string)
        self.check_splitinfo_format(entry.bunit_split_string)
        self.check_splitinfo_format(entry.cunit_split_string)

        if cols[14] == 'A' and \
                not (entry.aunit_split_string == '*' and entry.bunit_split_string == '*'):
            raise ValueError('invalid splitting')

        head_length = len(cols[0].encode('utf-8'))
        dict_from_wordid = -1 if cols[13] == '*' else int(cols[13])
        entry.wordinfo = WordInfo(
            cols[4], head_length, pos_id, cols[12], dict_from_wordid, '', cols[11], None, None, None)
        return entry

    def __is_length_valid(self, cols):
        head_length = len(cols[0].encode('utf-8'))
        return head_length <= self.__STRING_MAX_LENGTH \
            and len(cols[4]) <= self.__STRING_MAX_LENGTH \
            and len(cols[11]) <= self.__STRING_MAX_LENGTH \
            and len(cols[12]) <= self.__STRING_MAX_LENGTH

    def add_to_trie(self, headword, word_id):
        key = headword.encode('utf-8')
        if key not in self.trie_keys:
            self.trie_keys[key] = []
        self.trie_keys[key].append(word_id)

    def get_posid(self, strs):
        return self.pos_table.get_id(','.join(strs))

    def write_grammar(self, matrix_input_stream, output_stream):
        self.logger.info('writing the POS table...')
        self.convert_postable(self.pos_table.get_list())
        self.byte_buffer.seek(0)
        output_stream.write(self.byte_buffer.read())
        self.__logging_size(self.byte_buffer.tell())
        self.byte_buffer.clear()

        self.logger.info('writing the connection matrix...')
        if not matrix_input_stream:
            self.byte_buffer.write_int(0, 'short')
            self.byte_buffer.write_int(0, 'short')
            self.byte_buffer.seek(0)
            output_stream.write(self.byte_buffer.read())
            self.__logging_size(self.byte_buffer.tell())
            self.byte_buffer.clear()
            return
        matrix = self.convert_matrix(matrix_input_stream)
        self.byte_buffer.seek(0)
        output_stream.write(self.byte_buffer.read())
        self.byte_buffer.clear()
        output_stream.write(matrix.read())
        self.__logging_size(matrix.tell() + 4)

    def convert_postable(self, pos_list):
        self.byte_buffer.write_int(len(pos_list), 'short')
        for pos in pos_list:
            for text in pos.split(','):
                self.write_string(text)

    def convert_matrix(self, matrix_input):
        header = matrix_input.readline().strip()
        if re.fullmatch(r"\s*", header):
            raise ValueError('invalid format at line 0')
        lr = header.split()
        lsize, rsize = [int(x) for x in lr]
        self.byte_buffer.write_int(lsize, 'short')
        self.byte_buffer.write_int(rsize, 'short')

        matrix = JTypedByteBuffer()

        for i, line in enumerate(matrix_input.readlines()):
            line = line.strip()
            if re.fullmatch(r"\s*", line) or re.match("#", line):
                continue
            cols = line.split()
            if len(cols) < 3:
                self.logger.warn('invalid format at line {}'.format(i))
                continue
            l, r, cost = [int(col) for col in cols]
            pos = matrix.tell()
            matrix.seek(2 * (l + lsize * r))
            matrix.write_int(cost, 'short')
            matrix.seek(pos)
        return matrix

    def write_lexicon(self, io_out):
        trie = DoubleArray()
        wordid_table = JTypedByteBuffer()
        keys = []
        vals = []
        for key, word_ids in self.trie_keys.items():
            keys.append(key)
            vals.append(wordid_table.tell())
            wordid_table.write_int(len(word_ids), 'byte')
            for wid in word_ids:
                wordid_table.write_int(wid, 'int')

        self.logger.info('building the trie...')

        def progress_func(n, s):
            if (n % (s / 10 + 1)) == 0:
                self.logger('.')
        trie.build(keys, vals, progress_func)
        self.logger.info('done\n')

        self.logger.info('writing the trie...')
        self.byte_buffer.clear()
        self.byte_buffer.write_int(trie.size, 'int')
        self.byte_buffer.seek(0)
        io_out.write(self.byte_buffer.read())
        self.byte_buffer.clear()

        io_out.write(trie.byte_array().read())
        self.__logging_size(trie.size * 4 + 4)
        del trie

        self.logger.info('writing the word-ID table...')
        self.byte_buffer.write_int(wordid_table.tell(), 'int')
        self.byte_buffer.seek(0)
        io_out.write(self.byte_buffer.read())
        self.byte_buffer.clear()

        wordid_table.seek(0)
        io_out.write(wordid_table.read())
        self.__logging_size(wordid_table.tell() + 4)
        del wordid_table

        self.logger.info('writing the word parameters...')
        self.byte_buffer.write_int(len(self.entries), 'int')
        for entry in self.entries:
            self.byte_buffer.write_int(entry.parameters[0], 'short')
            self.byte_buffer.write_int(entry.parameters[1], 'short')
            self.byte_buffer.write_int(entry.parameters[2], 'short')
            self.byte_buffer.seek(0)
            io_out.write(self.byte_buffer.read())
            self.byte_buffer.clear()
        self.__logging_size(len(self.entries) * 6 + 4)
        self.write_wordinfo(io_out)

    def write_wordinfo(self, io_out):
        mark = io_out.tell()
        io_out.seek(mark * 4 + len(self.entries))
        offsets = JTypedByteBuffer()
        self.logger.info('writing the word_infos...')
        base = io_out.tell()
        for entry in self.entries:
            wi = entry.wordinfo
            offsets.write_int(io_out.tell(), 'int')
            self.write_string(wi.surface)
            self.write_stringlength(wi.length())
            self.byte_buffer.write_int(wi.pos_id, 'short')
            if wi.normalized_form == wi.surface:
                self.write_string('')
            else:
                self.write_string(wi.normalized_form)
            self.byte_buffer.write_int(wi.dictionary_form_word_id, 'int')
            if wi.reading_form == wi.surface:
                self.write_string('')
            else:
                self.write_string(wi.reading_form)

            self.write_intarray(self.parse_splitinfo(entry.aunit_split_string))
            self.write_intarray(self.parse_splitinfo(entry.bunit_split_string))
            self.write_intarray(self.parse_splitinfo(entry.cunit_split_string))
            self.byte_buffer.seek(0)
            io_out.write(self.byte_buffer.read())
            self.byte_buffer.clear()
        self.__logging_size(io_out.tell() - base)
        self.logger.info('writing word_info offsets...')
        io_out.seek(mark)
        offsets.seek(0)
        io_out.write(offsets.read())
        self.__logging_size(offsets.tell())

    def decode(self, str_):
        def replace(match):
            uni_text = match.group()
            uni_text = uni_text.replace('{', '').replace('}', '')
            if len(uni_text) > 6:
                uni_text = ('\\U000{}'.format(uni_text[2:]))
            return uni_text.encode('ascii').decode('unicode-escape')
        return re.sub(self.__PATTERN_UNICODE_LITERAL, replace, str_)

    def check_splitinfo_format(self, str_):
        if str_.count('/') + 1 > self.__ARRAY_MAX_LENGTH:
            raise ValueError('too many units')

    def parse_splitinfo(self, info):
        if info == '*':
            return []
        words = info.split('/')
        if len(words) > self.__ARRAY_MAX_LENGTH:
            raise ValueError('too many units')
        ids = []
        for word in words:
            if self.__is_id(word):
                ids.append(self.parse_id(word))
            else:
                ids.append(self.word_to_id(word))
                if ids[-1] < 0:
                    return ValueError('not found such a word')
        return ids

    @staticmethod
    def __is_id(text):
        return re.match(r'U?\d+', text)

    def parse_id(self, text):
        if text.startswith('U'):
            id_ = int(text[1:])
            if self.is_user_dictionary:
                id_ |= (1 << 28)
        else:
            id_ = int(text)
        self.check_wordid(id_)
        return id_

    def word_to_id(self, text):
        cols = text.split(',')
        if len(cols) < 8:
            raise ValueError('too few columns')
        headword = self.decode(cols[0])
        pos_id = self.get_posid([cols[i] for i in range(1, 7)])
        if pos_id < 0:
            raise ValueError('invalid part of speech')
        reading = self.decode(cols[7])
        return self.get_wordid(headword, pos_id, reading)

    def get_wordid(self, headword, pos_id, reading_form):
        for i in range(len(self.entries)):
            info = self.entries[i].wordinfo
            if info.surface == headword \
                    and info.pos_id == pos_id \
                    and info.reading_form == reading_form:
                return i
        return -1

    def check_wordid(self, wid):
        if wid < 0 or wid >= len(self.entries):
            raise ValueError('invalid word ID')

    def write_string(self, text):
        len_ = 0
        for c in text:
            if 0x10000 <= ord(c) <= 0x10FFFF:
                len_ += 2
            else:
                len_ += 1
        self.write_stringlength(len_)
        self.byte_buffer.write_str(text)

    def write_stringlength(self, len_):
        if len_ <= self.__BYTE_MAX_VALUE:
            self.byte_buffer.write_int(len_, 'byte')
        else:
            self.byte_buffer.write_int((len_ >> 8) | 0x80, 'byte')
            self.byte_buffer.write_int((len_ & 0xFF), 'byte')

    def write_intarray(self, array):
        self.byte_buffer.write_int(len(array), 'byte')
        for item in array:
            self.byte_buffer.write_int(item, 'int')

    def __logging_size(self, size):
        self.logger.info('{} bytes\n'.format(size))



from sudachipy import config
from sudachipy.dictionarylib.grammar import Grammar

from .editconnectioncost import EditConnectionCostPlugin


class InhibitConnectionPlugin(EditConnectionCostPlugin):

    def __init__(self):
        self.inhibit_pairs = []

    def set_up(self, grammar: Grammar) -> None:
        if 'inhibitedPair' in config.settings:
            self.inhibit_pairs = config.settings['inhibitedPair']

    def edit(self, grammar: Grammar) -> None:
        for pair in self.inhibit_pairs:
            if len(pair) < 2:
                continue
            self.inhibit_connection(grammar, pair[0], pair[1])



import struct

from sudachipy.dictionarylib.jtypedbytebuffer import JTypedByteBuffer


class DictionaryHeader:

    __DESCRIPTION_SIZE = 256
    __STORAGE_SIZE = 8 + 8 + __DESCRIPTION_SIZE

    def __init__(self, version, create_time, description):
        self.version = version
        self.create_time = create_time
        self.description = description

    @classmethod
    def from_bytes(cls, bytes_, offset):
        version, create_time = struct.unpack_from("<2Q", bytes_, offset)
        offset += 16

        len_ = 0
        while len_ < cls.__DESCRIPTION_SIZE:
            if bytes_[offset + len_] == 0:
                break
            len_ += 1
        description = bytes_[offset:offset + len_].decode("utf-8")
        return cls(version, create_time, description)

    def to_bytes(self):
        buf = JTypedByteBuffer(b'\x00' * (16 + self.__DESCRIPTION_SIZE))
        buf.seek(0)
        buf.write_int(self.version, 'long', signed=False)
        buf.write_int(self.create_time, 'long')
        bdesc = self.description.encode('utf-8')
        if len(bdesc) > self.__DESCRIPTION_SIZE:
            raise ValueError('description is too long')
        buf.write(bdesc)
        return buf.getvalue()

    def storage_size(self):
        return self.__STORAGE_SIZE



class DoubleArrayBuilderUnit(object):
    def __init__(self):
        self.unit = 0

    def set_has_leaf(self, has_leaf):
        """
        :param has_leaf: bool
        """
        if has_leaf:
            self.unit |= 1 << 8
        else:
            self.unit &= ~(1 << 8)

    def set_value(self, value):
        """
        :param value: int
        """
        self.unit = value | (1 << 31)

    def set_label(self, label):
        """
        :param label: bytes
        """
        self.unit = (self.unit & ~0xFF) | int(label)

    def set_offset(self, offset):
        self.unit &= (1 << 31) | (1 << 8) | 0xFF
        if offset < 1 << 21:
            self.unit |= (offset << 10)
        else:
            self.unit |= (offset << 2) | (1 << 9)



#from sudachipy.plugin.oov import simple_oov_plugin
#from sudachipy.plugin.oov import mecab_oov_plugin
#from sudachipy.plugin.path_rewrite import join_numeric_plugin
#from sudachipy.plugin.path_rewrite import join_katakana_oov_plugin
#from sudachipy.plugin.input_text import default_input_text_plugin
#from .connect_cost.inhibitconnectioncost import InhibitConnectionPlugin
#from . import oov
#from . import input_text
#from . import path_rewrite



from .dictionarybuilder import DictionaryBuilder


class UserDictionaryBuilder(DictionaryBuilder):

    def __init__(self, grammar, system_lexicon, *, logger=None):
        super().__init__(logger=logger)
        self.is_user_dictionary = True
        self.grammar = grammar
        self.system_lexicon = system_lexicon

    def build(self, lexicon_paths, matrix_input_stream, out_stream):
        """
        Violated LSP
        :param lexicon_paths:
        :param out_stream:
        :return:
        """
        self.logger.info('reading the source file...')
        for path in lexicon_paths:
            with open(path) as rf:
                self.build_lexicon(rf)
        self.logger.info('{} words\n'.format(len(self.entries)))

        self.write_grammar(None, out_stream)
        self.write_lexicon(out_stream)

    def get_posid(self, strs):
        pos_id = self.grammar.get_part_of_speech_id(strs)
        if pos_id < 0:
            pos_id = super().get_posid(strs) + self.grammar.get_part_of_speech_size()
        return pos_id

    def get_wordid(self, headword, pos_id, reading_form):
        wid = super().get_wordid(headword, pos_id, reading_form)
        if wid >= 0:
            return wid | (1 << 28)
        return self.system_lexicon.get_word_id(headword, pos_id, reading_form)

    def check_wordid(self, wid):
        if wid >= (1 << 28):
            super().check_wordid(wid & ((1 << 28) - 1))
        elif wid < 0 or wid >= self.system_lexicon.size():
            raise ValueError('invalid word id')



from abc import ABC, abstractmethod
from typing import Iterator, List

from .wordinfo import WordInfo


class Lexicon(ABC):

    Itr = Iterator[List[int]]

    @abstractmethod
    def lookup(self, text: str, offset: int) -> Itr:  # noqa: F821
        raise NotImplementedError

    @abstractmethod
    def get_word_id(self, headword: str, pos_id: int, reading_form: str) -> int:
        raise NotImplementedError

    @abstractmethod
    def get_left_id(self, word_id: int) -> int:
        raise NotImplementedError

    @abstractmethod
    def get_right_id(self, word_id: int) -> int:
        raise NotImplementedError

    @abstractmethod
    def get_cost(self, word_id: int) -> int:
        raise NotImplementedError

    @abstractmethod
    def get_word_info(self, word_id: int) -> 'WordInfo':
        raise NotImplementedError

    @abstractmethod
    def get_dictionary_id(self, word_id: int) -> int:
        raise NotImplementedError

    @abstractmethod
    def size(self) -> int:
        raise NotImplementedError



from . import latticenode
from . import morpheme
from . import tokenizer


class MorphemeList:
    def __init__(self, input_, grammar, lexicon, path):
        self.input_text = input_
        self.grammar = grammar
        self.lexicon = lexicon
        self.path = path

    def __getitem__(self, index):
        return morpheme.Morpheme(self, index)

    def __len__(self):
        return len(self.path)

    def __iter__(self):
        for index in range(len(self.path)):
            yield morpheme.Morpheme(self, index)
        return

    def get_begin(self, index):
        return self.input_text.get_original_index(self.path[index].get_begin())

    def get_end(self, index):
        return self.input_text.get_original_index(self.path[index].get_end())

    def get_surface(self, index):
        begin = self.get_begin(index)
        end = self.get_end(index)
        return self.input_text.get_original_text()[begin:end]

    def get_word_info(self, index):
        return self.path[index].get_word_info()

    def split(self, mode, index, wi):
        if mode is tokenizer.Tokenizer.SplitMode.A:
            word_ids = wi.a_unit_split
        elif mode is tokenizer.Tokenizer.SplitMode.B:
            word_ids = wi.b_unit_split
        else:
            return [self.__getitem__(index)]

        if len(word_ids) is 0 or len(word_ids) is 1:
            return [self.__getitem__(index)]

        offset = self.path[index].get_begin()
        nodes = []
        for wid in word_ids:
            n = latticenode.LatticeNode(self.lexicon, 0, 0, 0, wid)
            n.begin = offset
            offset += n.get_word_info().head_word_length
            n.end = offset
            nodes.append(n)

        return MorphemeList(self.input_text, self.grammar, self.lexicon, nodes)

    def is_oov(self, index):
        return self.path[index].is_oov

    def get_internal_cost(self):
        return self.path[-1].get_path_cost() - self.path[0].get_path_cost()

    def size(self):
        return len(self.path)



from sudachipy import config

from . import InputTextPlugin


class ProlongedSoundMarkInputTextPlugin(InputTextPlugin):

    def __init__(self):
        self.psm_set = set()
        self.replace_symbol = ''

    def set_up(self) -> None:
        for s in config.settings['prolongedSoundMarks']:
            self.psm_set.add(ord(s[0]))
        self.replace_symbol = config.settings['replacementSymbol']

    def rewrite(self, builder: InputTextPlugin.Builder) -> None:
        text = builder.get_text()
        n = len(text)
        offset = 0
        is_psm = False
        m_start_idx = n
        for i in range(n):
            cp = ord(text[i])
            if not is_psm and cp in self.psm_set:
                is_psm = True
                m_start_idx = i
            elif is_psm and cp not in self.psm_set:
                if (i - m_start_idx) > 1:
                    builder.replace(m_start_idx - offset, i - offset, self.replace_symbol)
                    offset += i - m_start_idx - 1
                is_psm = False
        if is_psm and (n - m_start_idx) > 1:
            builder.replace(m_start_idx - offset, n - offset, self.replace_symbol)



from typing import List

from sudachipy import config

from . import MeCabOovPlugin, OovProviderPlugin, SimpleOovPlugin


def get_oov_plugin(json_obj) -> OovProviderPlugin:
    # In the future, users can define plugin by themselves
    try:
        if json_obj['class'] == 'sudachipy.plugin.oov.MeCabOovProviderPlugin':
            return MeCabOovPlugin(json_obj)
        if json_obj['class'] == 'sudachipy.plugin.oov.SimpleOovProviderPlugin':
            return SimpleOovPlugin(json_obj)
        raise ValueError('{} is invalid OovProviderPlugin class'.format(json_obj['class']))
    except KeyError:
        raise ValueError('config file is invalid format')


def get_oov_plugins() -> List[OovProviderPlugin]:
    key_word = 'oovProviderPlugin'
    if key_word not in config.settings:
        return []
    ps = []
    for obj in config.settings[key_word]:
        ps.append(get_oov_plugin(obj))
    return ps



class WordParameterList(object):
    def __init__(self, bytes_, offset):
        self.ELEMENT_SIZE = 2 * 3

        self.bytes = bytes_
        bytes_.seek(offset)
        self.size = int.from_bytes(bytes_.read(4), 'little')
        self.offset = offset + 4
        # self.is_copied = False

    def storage_size(self):
        return 4 + self.ELEMENT_SIZE * self.size

    def get_size(self):
        return self.size

    def get_left_id(self, word_id):
        self.bytes.seek(self.offset + self.ELEMENT_SIZE * word_id)
        return int.from_bytes(self.bytes.read(2), 'little', signed=True)

    def get_right_id(self, word_id):
        self.bytes.seek(self.offset + self.ELEMENT_SIZE * word_id + 2)
        return int.from_bytes(self.bytes.read(2), 'little')

    def get_cost(self, word_id):
        self.bytes.seek(self.offset + self.ELEMENT_SIZE * word_id + 4)
        return int.from_bytes(self.bytes.read(2), 'little', signed=True)

    def set_cost(self, word_id, cost):
        # bytes_ must be ACCESS_COPY mode
        self.bytes.seek(self.offset + self.ELEMENT_SIZE * word_id + 4)
        self.bytes.write(cost.to_bytes(2, 'little', signed=True))



class Morpheme:
    def __init__(self, list_, index):
        self.word_info = None
        self.list = list_
        self.index = index

    def begin(self):
        return self.list.get_begin(self.index)

    def end(self):
        return self.list.get_end(self.index)

    def surface(self):
        return self.list.get_surface(self.index)

    def part_of_speech(self):
        wi = self.get_word_info()
        return self.list.grammar.get_part_of_speech_string(wi.pos_id)

    def part_of_speech_id(self):
        wi = self.get_word_info()
        return wi.pos_id

    def dictionary_form(self):
        wi = self.get_word_info()
        return wi.dictionary_form

    def normalized_form(self):
        wi = self.get_word_info()
        return wi.normalized_form

    def reading_form(self):
        wi = self.get_word_info()
        return wi.reading_form

    def split(self, mode):
        wi = self.get_word_info()
        return self.list.split(mode, self.index, wi)

    def is_oov(self):
        return self.list.is_oov(self.index)

    def word_id(self):
        return self.list.path[self.index].get_word_id()

    def dictionary_id(self):
        return self.list.path[self.index].get_dictionary_id()

    def get_word_info(self):
        if not self.word_info:
            self.word_info = self.list.get_word_info(self.index)
        return self.word_info



import argparse
import fileinput
import os
import sys
import time

from . import dictionary
from . import tokenizer
from .dictionarylib import BinaryDictionary
from .dictionarylib import SYSTEM_DICT_VERSION, USER_DICT_VERSION_2
from .dictionarylib.dictionarybuilder import DictionaryBuilder
from .dictionarylib.dictionaryheader import DictionaryHeader
from .dictionarylib.userdictionarybuilder import UserDictionaryBuilder


def _set_default_subparser(self, name, args=None):
    """
    copy and modify code from https://bitbucket.org/ruamel/std.argparse
    """
    subparser_found = False
    for arg in sys.argv[1:]:
        if arg in ['-h', '--help']:  # global help if no subparser
            break
    else:
        for x in self._subparsers._actions:
            if not isinstance(x, argparse._SubParsersAction):
                continue
            for sp_name in x._name_parser_map.keys():
                if sp_name in sys.argv[1:]:
                    subparser_found = True
        if not subparser_found:
            # insert default in first position, this implies no
            # global options without a sub_parsers specified
            if args is None:
                sys.argv.insert(1, name)
            else:
                args.insert(0, name)


argparse.ArgumentParser.set_default_subparser = _set_default_subparser


def run(tokenizer, mode, input, output, print_all):
    for line in input:
        line = line.rstrip('\n')
        for m in tokenizer.tokenize(line, mode):
            list_info = [
                m.surface(),
                ",".join(m.part_of_speech()),
                m.normalized_form()]
            if print_all:
                list_info += [
                    m.dictionary_form(),
                    m.reading_form(),
                    str(m.dictionary_id())]
                if m.is_oov():
                    list_info.append("(OOV)")
            print("\t".join(list_info), file=output)
        print("EOS", file=output)


def _system_dic_checker(args, print_usage):
    if not args.system_dic:
        args.system_dic = os.path.join(os.path.dirname(
            os.path.abspath(__file__)), os.pardir, 'resources/system.dic')
    if not os.path.exists(args.system_dic):
        print_usage()
        print('{}: error: {} doesn\'t exist'.format(__name__, args.system_dic))
        exit()


def _input_files_checker(args, print_usage):
    if not args.in_files:
        print_usage()
        print('{}: error: no input files'.format(__name__))
        exit()
    for file in args.in_files:
        if not os.path.exists(file):
            print_usage()
            print('{}: error: {} doesn\'t exist'.format(__name__, file))
            exit()


def _matrix_file_checker(args, print_usage):
    if not os.path.exists(args.matrix_file):
        print_usage()
        print('{}: error: {} doesn\'t exist'.format(__name__, args.matrix_file))
        exit()


def _command_user_build(args, print_usage):
    _system_dic_checker(args, print_usage)
    _input_files_checker(args, print_usage)
    header = DictionaryHeader(
        USER_DICT_VERSION_2, int(time.time()), args.description)
    dict_ = BinaryDictionary.from_system_dictionary(args.system_dic)
    with open(args.out_file, 'wb') as wf:
        wf.write(header.to_bytes())
        builder = UserDictionaryBuilder(dict_.grammar, dict_.lexicon)
        builder.build(args.in_files, None, wf)


def _command_build(args, print_usage):
    _matrix_file_checker(args, print_usage)
    _input_files_checker(args, print_usage)
    header = DictionaryHeader(
        SYSTEM_DICT_VERSION, int(time.time()), args.description)
    with open(args.out_file, 'wb') as wf, open(args.matrix_file, 'r') as rf:
        wf.write(header.to_bytes())
        builder = DictionaryBuilder()
        builder.build(args.in_files, rf, wf)


def _command_tokenize(args, print_usage):
    _input_files_checker(args, print_usage)

    if args.mode == "A":
        mode = tokenizer.Tokenizer.SplitMode.A
    elif args.mode == "B":
        mode = tokenizer.Tokenizer.SplitMode.B
    else:
        mode = tokenizer.Tokenizer.SplitMode.C

    output = sys.stdout
    if args.fpath_out:
        output = open(args.fpath_out, "w", encoding="utf-8")

    print_all = args.a

    is_enable_dump = args.d

    dict_ = dictionary.Dictionary(config_path=args.fpath_setting)
    tokenizer_obj = dict_.create()
    if is_enable_dump:
        tokenizer_obj.set_dump_output(output)

    input_ = fileinput.input(args.in_files, openhook=fileinput.hook_encoded("utf-8"))
    run(tokenizer_obj, mode, input_, output, print_all)

    output.close()


def main():
    parser = argparse.ArgumentParser(description="Japanese Morphological Analyzer")

    subparsers = parser.add_subparsers(description='')

    parser.add_argument("-v", "--version", action="version", version="%(prog)s v0.2.0")

    # root, tokenizer parser
    parser_tk = subparsers.add_parser('tokenize', help='(default) see `tokenize -h`', description='Tokenize Text')
    parser_tk.add_argument("-r", dest="fpath_setting", metavar="file", help="the setting file in JSON format")
    parser_tk.add_argument("-m", dest="mode", choices=["A", "B", "C"], default="C", help="the mode of splitting")
    parser_tk.add_argument("-o", dest="fpath_out", metavar="file", help="the output file")
    parser_tk.add_argument("-a", action="store_true", help="print all of the fields")
    parser_tk.add_argument("-d", action="store_true", help="print the debug information")
    parser_tk.add_argument("in_files", metavar="file", nargs=argparse.ONE_OR_MORE, help='text written in utf-8')
    parser_tk.set_defaults(handler=_command_tokenize, print_usage=parser_tk.print_usage)

    # build dictionary parser
    parser_bd = subparsers.add_parser('build', help='see `build -h`', description='Build Sudachi Dictionary')
    parser_bd.add_argument('-o', dest='out_file', metavar='file', default='system.dic',
                           help='output file (default: system.dic)')
    parser_bd.add_argument('-d', dest='description', default='', metavar='string', required=False,
                           help='description comment to be embedded on dictionary')
    required_named_bd = parser_bd.add_argument_group('required named arguments')
    required_named_bd.add_argument('-m', dest='matrix_file', metavar='file', required=True,
                                   help='connection matrix file with MeCab\'s matrix.def format')
    parser_bd.add_argument("in_files", metavar="file", nargs=argparse.ONE_OR_MORE,
                           help='source files with CSV format (one of more)')
    parser_bd.set_defaults(handler=_command_build, print_usage=parser_bd.print_usage)

    # build user-dictionary parser
    parser_ubd = subparsers.add_parser('ubuild', help='see `ubuild -h`', description='Build User Dictionary')
    parser_ubd.add_argument('-d', dest='description', default='', metavar='string', required=False,
                            help='description comment to be embedded on dictionary')
    parser_ubd.add_argument('-o', dest='out_file', metavar='file', default='user.dic',
                            help='output file (default: user.dic)')
    parser_ubd.add_argument('-s', dest='system_dic', metavar='file', required=False,
                            help='system dictionary (default: ${SUDACHIPY}/resouces/system.dic)')
    parser_ubd.add_argument("in_files", metavar="file", nargs=argparse.ONE_OR_MORE,
                            help='source files with CSV format (one or more)')
    parser_ubd.set_defaults(handler=_command_user_build, print_usage=parser_ubd.print_usage)

    parser.set_default_subparser('tokenize')

    args = parser.parse_args()

    if hasattr(args, 'handler'):
        args.handler(args, args.print_usage)
    else:
        parser.print_help()


# Todo: delete this function in the future
def _read_system_dictionary(filename):
    from .dictionarylib.dictionaryheader import DictionaryHeader
    from . import dictionarylib
    """
    Copy of sudachipy.dictionary.Dictionary.read_system_dictionary
    :param filename:
    :return:
    """
    import mmap
    buffers = []
    if filename is None:
        raise AttributeError("system dictionary is not specified")
    with open(filename, 'r+b') as system_dic:
        bytes_ = mmap.mmap(system_dic.fileno(), 0, access=mmap.ACCESS_READ)
    buffers.append(bytes_)

    offset = 0
    header = DictionaryHeader.from_bytes(bytes_, offset)
    if header.version != SYSTEM_DICT_VERSION:
        raise Exception("invalid system dictionary")
    offset += header.storage_size()

    grammar = dictionarylib.grammar.Grammar(bytes_, offset)
    offset += grammar.get_storage_size()

    lexicon = dictionarylib.lexiconset.LexiconSet(dictionarylib.doublearraylexicon.DoubleArrayLexicon(bytes_, offset))
    return buffers, header, grammar, lexicon


if __name__ == '__main__':
    main()



import mmap
import struct

from . import wordidtable
from . import wordinfolist
from . import wordparameterlist
from .lexicon import Lexicon
from .. import dartsclone


class DoubleArrayLexicon(Lexicon):

    __SIGNED_SHORT_MIN = -32768
    __SIGNED_SHORT_MAX = 32767
    __USER_DICT_COST_PER_MORPH = -20

    def __init__(self, bytes_: mmap.mmap, offset: int):
        self.trie = dartsclone.doublearray.DoubleArray()
        bytes_.seek(offset)
        size = int.from_bytes(bytes_.read(4), 'little')
        offset += 4
        # bytes_.seek(offset)
        array = struct.unpack_from("<{}I".format(size), bytes_, offset)
        self.trie.set_array(array, size)
        offset += self.trie.total_size()

        self.word_id_table = wordidtable.WordIdTable(bytes_, offset)
        offset += self.word_id_table.storage_size()

        self.word_params = wordparameterlist.WordParameterList(bytes_, offset)
        offset += self.word_params.storage_size()

        self.word_infos = wordinfolist.WordInfoList(bytes_, offset, self.word_params.get_size())

    def lookup(self, text: bytes, offset: int) -> Lexicon.Itr:
        result = self.trie.common_prefix_search(text, offset)
        for item in result:
            word_ids = self.word_id_table.get(item[0])
            length = item[1]
            for word_id in word_ids:
                yield (word_id, length)

    def get_left_id(self, word_id: int) -> int:
        return self.word_params.get_left_id(word_id)

    def get_right_id(self, word_id: int) -> int:
        return self.word_params.get_right_id(word_id)

    def get_cost(self, word_id: int) -> int:
        return self.word_params.get_cost(word_id)

    def get_word_info(self, word_id: int) -> 'WordInfo':  # noqa: F821
        return self.word_infos.get_word_info(word_id)

    def size(self) -> int:
        return self.word_params.size

    def get_word_id(self, headword: str, pos_id: int, reading_form: str) -> int:
        for wid in range(self.word_infos.size()):
            info = self.word_infos.get_word_info(wid)
            if info.surface == headword \
                    and info.pos_id == pos_id \
                    and info.reading_form == reading_form:
                return wid
        return -1

    def get_dictionary_id(self, word_id: int) -> int:
        return 0

    def calculate_cost(self, tokenizer) -> None:
        for wid in range(self.word_params.size):
            if self.get_cost(wid) != self.__SIGNED_SHORT_MIN:
                continue
            surface = self.get_word_info(wid).surface
            ms = tokenizer.tokenize(surface, None)
            cost = ms.get_internal_cost() + self.__USER_DICT_COST_PER_MORPH * len(ms)
            cost = min(cost, self.__SIGNED_SHORT_MAX)
            cost = max(cost, self.__SIGNED_SHORT_MIN)
            self.word_params.set_cost(wid, cost)



from . import grammar
from . import charactercategory
from . import categorytype
from . import lexiconset
from . import doublearraylexicon
from . import dictionaryheader
from .dictionaryversion import SYSTEM_DICT_VERSION, USER_DICT_VERSION_1, USER_DICT_VERSION_2
from .binarydictionary import BinaryDictionary



class BitVector(object):
    def __init__(self):
        self.UNIT_SIZE = 32
        self.units = []
        self.ranks = []
        self.num_ones = 0
        self.size = 0

    def get(self, id):
        return (self.units[int(id / self.UNIT_SIZE)] >> (id % self.UNIT_SIZE) & 1) is 1

    def rank(self, id):
        unit_id = int(id / self.UNIT_SIZE)
        return self.ranks[unit_id] \
            + self.pop_count(self.units[unit_id] & (~0 >> (self.UNIT_SIZE - (id % self.UNIT_SIZE) - 1)))

    def set(self, id, bit):
        if bit:
            self.units[int(id / self.UNIT_SIZE)] \
                = self.units[int(id / self.UNIT_SIZE)] | 1 << (id % self.UNIT_SIZE)
        else:
            self.units[int(id / self.UNIT_SIZE)] \
                = self.units[int(id / self.UNIT_SIZE)] & ~(1 << (id % self.UNIT_SIZE))

    def is_empty(self):
        return len(self.units) is 0

    def get_num_ones(self):
        return self.num_ones

    def get_size(self):
        return self.size

    def append(self):
        if (self.size % self.UNIT_SIZE) is 0:
            self.units.append(0)
        self.size += 1

    def build(self):
        self.ranks = [0] * len(self.units)
        self.num_ones = 0
        for i in range(len(self.units)):
            self.ranks[i] = self.num_ones
            self.num_ones += self.pop_count(self.units[i])

    def clear(self):
        self.units.clear()
        self.ranks = None

    def pop_count(self, unit):
        unit = ((unit & 0xAAAAAAAA) >> 1) + (unit & 0x55555555)
        unit = ((unit & 0xCCCCCCCC) >> 2) + (unit & 0x33333333)
        unit = ((unit >> 4) + unit) & 0x0F0F0F0F
        unit += unit >> 8
        unit += unit >> 16
        return unit & 0xFF



from sudachipy.dictionarylib import wordinfo

from . import OovProviderPlugin


class SimpleOovPlugin(OovProviderPlugin):

    def __init__(self, json_obj):
        self.left_id = json_obj['leftId']
        self.right_id = json_obj['rightId']
        self.cost = json_obj['cost']
        self.__oov_pos_strings = json_obj['oovPOS']
        self.oov_pos_id = -1

    def set_up(self, grammar):
        self.oov_pos_id = grammar.get_part_of_speech_id(self.__oov_pos_strings)

    def provide_oov(self, input_text, offset, has_other_words):
        if not has_other_words:
            node = self.create_node()
            node.set_parameter(self.left_id, self.right_id, self.cost)
            length = input_text.get_code_points_offset_length(offset, 1)
            s = input_text.get_substring(offset, offset + length)
            info = wordinfo.WordInfo(surface=s, head_word_length=length, pos_id=self.oov_pos_id, normalized_form=s,
                                     dictionary_form_word_id=-1, dictionary_form=s, reading_form="",
                                     a_unit_split=[], b_unit_split=[], word_structure=[])
            node.set_word_info(info)
            return [node]
        else:
            return []



from typing import List

from sudachipy import config

from . import JoinKatakanaOovPlugin, JoinNumericPlugin, PathRewritePlugin


def get_path_rewrite_plugin(json_obj) -> PathRewritePlugin:
    # In the future, users can define plugin by themselves
    try:
        if json_obj['class'] == 'sudachipy.plugin.path_rewrite.JoinNumericPlugin':
            return JoinNumericPlugin(json_obj)
        if json_obj['class'] == 'sudachipy.plugin.path_rewrite.JoinKatakanaOovPlugin':
            return JoinKatakanaOovPlugin(json_obj)
        raise ValueError('{} is invalid PathRewritePlugin class'.format(json_obj['class']))
    except KeyError:
        raise ValueError('config file is invalid format')


def get_path_rewrite_plugins() -> List[PathRewritePlugin]:
    if 'pathRewritePlugin' not in config.settings:
        return []
    ps = []
    for obj in config.settings['pathRewritePlugin']:
        ps.append(get_path_rewrite_plugin(obj))
    return ps



import copy

from . import bitvector


class DAWGBuilder(object):
    class Node(object):
        def __init__(self):
            self.child = 0
            self.sibling = 0
            self.label = 0
            self.is_state = False
            self.has_sibling = False

        def reset(self):
            self.child = 0
            self.sibling = 0
            self.label = 0
            self.is_state = False
            self.has_sibling = False

        def get_value(self):
            return self.child

        def set_value(self, value):
            self.child = value

        def unit(self):
            if self.label is 0:
                return (self.child << 1) | (1 if self.has_sibling else 0)
            return (self.child << 2) | (2 if self.is_state else 0) | (1 if self.has_sibling else 0)

        def __str__(self):
            return "child: %d, sibling: %d, label: %c, is_state: %d, has_sibling: %d" % (
                self.child, self.sibling, self.label, int(self.is_state), int(self.has_sibling)
            )

    class Unit(object):
        def __init__(self):
            self.unit = 0

        def child(self):
            return self.unit >> 2

        def has_sibling(self):
            return (self.unit & 1) is 1

        def value(self):
            return self.unit >> 1

        def is_state(self):
            return (self.unit & 2) is 2

    def __init__(self):
        self.INITIAL_TABLE_SIZE = 1 << 10

        self.nodes = []
        self.units = []
        self.labels = []
        self.is_intersections = bitvector.BitVector()
        self.table = []
        self.node_stack = []
        self.recycle_bin = []
        self.num_states = 0

    def root(self):
        return 0

    def child(self, id):
        return self.units[id].child()

    def sibling(self, id):
        return id + 1 if self.units[id].has_sibling() else 0

    def value(self, id):
        return self.units[id].value()

    def is_leaf(self, id):
        return self.label(id) is 0

    def label(self, id):
        return self.labels[id]

    def is_intersection(self, id):
        return self.is_intersections.get(id)

    def intersecrion_id(self, id):
        return self.is_intersections.rank(id) - 1

    def num_intersections(self):
        return self.is_intersections.get_num_ones()

    def size(self):
        return len(self.units)

    def init(self):
        self.table = [0] * self.INITIAL_TABLE_SIZE

        self.append_node()
        self.append_unit()

        self.num_states = 1

        self.nodes[0].label = 0xFF
        self.node_stack.append(0)

    def finish(self):
        self.flush(0)

        self.units[0].unit = copy.deepcopy(self.nodes[0].unit())
        self.labels[0] = copy.deepcopy(self.nodes[0].label)

        self.nodes = None
        self.table = None
        self.node_stack = None
        self.recycle_bin = None

        self.is_intersections.build()

    def insert(self, key, value):
        if value < 0:
            raise AttributeError("negative value")
        if len(key) is 0:
            raise AttributeError("zero-length key")

        id_ = 0
        key_pos = 0

        while key_pos <= len(key):
            child_id = self.nodes[id_].child
            if child_id is 0:
                break

            key_label = key[key_pos] if key_pos < len(key) else 0
            if key_pos < len(key) and key_label is 0:
                raise AttributeError("invalid null character")

            unit_label = self.nodes[child_id].label
            if key_label < unit_label:
                raise AttributeError("wrong key order")
            elif key_label > unit_label:
                self.nodes[child_id].has_sibling = True
                self.flush(child_id)
                break
            id_ = child_id
            key_pos += 1

        if key_pos > len(key):
            return

        while key_pos <= len(key):
            key_label = key[key_pos] if key_pos < len(key) else 0
            child_id = self.append_node()

            if self.nodes[id_].child is 0:
                self.nodes[child_id].is_state = True
            self.nodes[child_id].sibling = self.nodes[id_].child
            self.nodes[child_id].label = key_label
            self.nodes[id_].child = child_id
            self.node_stack.append(child_id)

            id_ = child_id
            key_pos += 1
        self.nodes[id_].set_value(value)

    def clear(self):

        self.nodes = []
        self.units = []
        self.labels = []
        self.is_intersections = bitvector.BitVector()
        self.table = []
        self.node_stack = []
        self.recycle_bin = []
        self.num_states = 0

    def flush(self, id):
        while self.stack_top(self.node_stack) is not id:
            node_id = self.stack_top(self.node_stack)
            self.stack_pop(self.node_stack)

            if self.num_states >= len(self.table) - (len(self.table) >> 2):
                self.expand_table()

            num_siblings = 0
            n = node_id
            while n is not 0:
                num_siblings += 1
                n = self.nodes[n].sibling

            find_result = self.find_node(node_id)
            match_id = find_result[0]
            hash_id = find_result[1]

            if match_id is not 0:
                self.is_intersections.set(match_id, True)
            else:
                unit_id = 0
                for i in range(num_siblings):
                    unit_id = self.append_unit()
                n = node_id
                while n is not 0:
                    self.units[unit_id].unit = self.nodes[n].unit()
                    self.labels[unit_id] = self.nodes[n].label
                    unit_id -= 1
                    n = self.nodes[n].sibling
                match_id = unit_id + 1
                self.table[hash_id] = match_id
                self.num_states += 1

            n = node_id
            while n is not 0:
                next_ = self.nodes[n].sibling
                self.free_node(n)
                n = next_

            self.nodes[self.stack_top(self.node_stack)].child = match_id
        self.stack_pop(self.node_stack)

    def expand_table(self):
        table_size = len(self.table) << 1
        self.table.clear()
        self.table = [0] * table_size

        for id_ in range(1, len(self.units)):
            if self.labels[id_] is 0 or self.units[id_].is_state():
                find_result = self.find_unit(id_)
                hash_id = find_result[1]
                self.table[hash_id] = id_

    def find_unit(self, id):
        result = [0] * 2
        hash_id = self.hash_unit(id) % len(self.table)
        while True:
            unit_id = self.table[hash_id]
            if unit_id is 0:
                break
            hash_id = (hash_id + 1) % len(self.table)
        result[1] = hash_id
        return result

    def find_node(self, node_id):
        result = [0] * 2
        hash_id = self.hash_node(node_id) % len(self.table)
        while True:
            unit_id = self.table[hash_id]
            if unit_id is 0:
                break

            if self.are_equal(node_id, unit_id):
                result[0] = unit_id
                result[1] = hash_id
                return result
            hash_id = (hash_id + 1) % len(self.table)

        result[1] = hash_id
        return result

    def are_equal(self, node_id, unit_id):
        n = self.nodes[node_id].sibling
        while n is not 0:
            if not self.units[unit_id].has_sibling():
                return False
            unit_id += 1
            n = self.nodes[n].sibling

        if self.units[unit_id].has_sibling():
            return False

        n = node_id
        while n is not 0:
            if (self.nodes[n].unit() is not self.units[unit_id].unit) or (self.nodes[n].label is not self.labels[unit_id]):
                return False
            n = self.nodes[n].sibling
            unit_id -= 1

        return True

    def hash_unit(self, id):
        hash_value = 0
        while id is not 0:
            unit = self.units[id].unit
            label = self.labels[id]
            hash_value ^= self.hash((label << 24) ^ unit)

            if not self.units[id].has_sibling():
                break
            id += 1
        return hash_value

    def hash_node(self, id_):
        hash_value = 0
        while id_ is not 0:
            unit = self.nodes[id_].unit()
            label = self.nodes[id_].label
            hash_value ^= self.hash(((label & 0xFF) << 24) ^ unit)
            id_ = self.nodes[id_].sibling
        return hash_value

    def append_unit(self):
        self.is_intersections.append()
        self.units.append(self.Unit())
        self.labels.append(0)

        return self.is_intersections.get_size() - 1

    def append_node(self):
        if len(self.recycle_bin) is 0:
            id_ = len(self.nodes)
            self.nodes.append(self.Node())
        else:
            id_ = self.stack_top(self.recycle_bin)
            self.nodes[id_].reset()
            self.stack_pop(self.recycle_bin)
        return id_

    def free_node(self, id):
        self.recycle_bin.append(id)

    def hash(self, key):
        key = ~key + (key << 15)  # key = (key << 15) - key - 1
        key = key ^ (key >> 12)
        key = key + (key << 2)
        key = key ^ (key >> 4)
        key = key * 2057  # key = (key + (key << 3)) + (key << 1)
        key = key ^ (key >> 16)
        return key

    def stack_top(self, stack):
        return stack[-1]

    def stack_pop(self, stack):
        stack.pop()

    def byte_to_u_int(self, b):
        return b & 0xFF



from typing import List

from .lexicon import Lexicon


class LexiconSet(Lexicon):

    __MAX_DICTIONARIES = 16

    def __init__(self, system_lexicon: Lexicon):
        self.lexicons = [system_lexicon]
        self.pos_offsets = []

    def add(self, lexicon: Lexicon, pos_offset: int) -> None:
        if lexicon not in self.lexicons:
            self.lexicons.append(lexicon)
            self.pos_offsets.append(pos_offset)

    def is_full(self) -> bool:
        return len(self.lexicons) >= self.__MAX_DICTIONARIES

    def lookup(self, text: str, offset: int) -> Lexicon.Itr:
        if len(self.lexicons) == 1:
            return self.lexicons[0].lookup(text, offset)
        return self.__lookup(text, offset)

    def __lookup(self, text: str, offset: int) -> Lexicon.Itr:
        indices = list(range(len(self.lexicons)))[1:] + [0]
        for dict_id in indices:
            pairs = self.lexicons[dict_id].lookup(text, offset)
            for pair in pairs:
                yield (self.build_word_id(dict_id, pair[0]), pair[1])

    def get_left_id(self, word_id: int) -> int:
        return self.lexicons[self.get_dictionary_id(word_id)]\
            .get_left_id(self.get_word_id1(word_id))

    def get_right_id(self, word_id: int) -> int:
        return self.lexicons[self.get_dictionary_id(word_id)]\
            .get_right_id(self.get_word_id1(word_id))

    def get_cost(self, word_id: int) -> int:
        return self.lexicons[self.get_dictionary_id(word_id)]\
            .get_cost(self.get_word_id1(word_id))

    def get_word_info(self, word_id: int) -> 'WordInfo':  # noqa: F821
        dic_id = self.get_dictionary_id(word_id)
        winfo = self.lexicons[dic_id].get_word_info(self.get_word_id1(word_id))
        pos_id = winfo.pos_id
        if dic_id > 0 and pos_id >= self.pos_offsets[1]:
            winfo.pos_id = winfo.pos_id - self.pos_offsets[1] + self.pos_offsets[dic_id]
        winfo.a_unit_split = self.convert_split(winfo.a_unit_split, dic_id)
        winfo.b_unit_split = self.convert_split(winfo.b_unit_split, dic_id)
        winfo.word_structure = self.convert_split(winfo.word_structure, dic_id)
        return winfo

    def get_dictionary_id(self, word_id: int) -> int:
        return word_id >> 28

    @staticmethod
    def get_word_id1(word_id: int) -> int:
        return 0x0FFFFFFF & word_id

    def get_word_id(self, headword: str, pos_id: int, reading_form: str) -> int:
        for dic_id in range(len(self.lexicons)):
            wid = self.lexicons[dic_id].get_word_id(headword, pos_id, reading_form)
            if wid <= 0:
                return self.build_word_id(dic_id, wid)
        return self.lexicons[0].get_word_id(headword, pos_id, reading_form)

    def build_word_id(self, dict_id, word_id):
        if word_id > 0x0FFFFFFF:
            raise AttributeError("word ID is too large: ", word_id)
        if dict_id > len(self.lexicons):
            raise AttributeError("dictionary ID is too large: ", dict_id)
        return (dict_id << 28) | word_id

    def size(self) -> int:
        return sum([lex.size() for lex in self.lexicons])

    def convert_split(self, split: List[int], dict_id: int) -> List[int]:
        for i in range(len(split)):
            if self.get_dictionary_id(split[i]) > 0:
                split[i] = self.build_word_id(dict_id, self.get_word_id1(split[i]))
        return split



from abc import ABC, abstractmethod
from typing import List

from sudachipy.dictionarylib.grammar import Grammar
from sudachipy.latticenode import LatticeNode
from sudachipy.utf8inputtext import UTF8InputText


class OovProviderPlugin(ABC):

    @abstractmethod
    def set_up(self, grammar: Grammar) -> None:
        raise NotImplementedError

    @abstractmethod
    def provide_oov(self, input_text: UTF8InputText, offset: int, has_other_words: bool) -> List[LatticeNode]:
        raise NotImplementedError

    def get_oov(self, input_text: UTF8InputText, offset: int, has_other_words: bool) -> List[LatticeNode]:
        nodes = self.provide_oov(input_text, offset, has_other_words)
        for node in nodes:
            node.begin = offset
            node.end = offset + node.get_word_info().length()
        return nodes

    @staticmethod
    def create_node() -> LatticeNode:
        node = LatticeNode()
        node.set_oov()
        return node



from typing import List, Optional

from .dictionarylib.grammar import Grammar
from .latticenode import LatticeNode


class Lattice:

    size = 0
    capacity = 0
    eos_node = None

    def __init__(self, grammar: Grammar):
        self.end_lists = []
        self.grammar = grammar
        self.eos_params = grammar.get_eos_parameter()
        bos_node = LatticeNode()
        bos_params = grammar.get_bos_parameter()
        bos_node.set_parameter(bos_params[0], bos_params[1], bos_params[2])
        bos_node.is_connected_to_bos = True
        self.end_lists.append([bos_node])

    def resize(self, size: int) -> None:
        if size > self.capacity:
            self.expand(size)
        self.size = size
        self.eos_node = LatticeNode()
        self.eos_node.set_parameter(self.eos_params[0], self.eos_params[1], self.eos_params[2])
        self.eos_node.begin = self.eos_node.end = size

    def clear(self) -> None:
        for i in range(1, self.size + 1):
            self.end_lists[i].clear()
        self.size = 0
        self.eos_node = None

    def expand(self, new_size: int) -> None:
        expand_list = [[] for _ in range(self.size, new_size)]
        self.end_lists.extend(expand_list)
        self.capacity = new_size

    def get_nodes_with_end(self, end: int) -> List[LatticeNode]:
        return self.end_lists[end]

    def get_nodes(self, begin: int, end: int) -> List[LatticeNode]:
        return [node for node in self.end_lists[end] if node.begin == begin]

    def get_minumum_node(self, begin: int, end: int) -> Optional[LatticeNode]:
        nodes = self.get_nodes(begin, end)
        if not nodes:
            return None
        min_arg = nodes[0]
        for node in nodes[1:]:
            if node.cost < min_arg.cost:
                min_arg = node
        return min_arg

    def insert(self, begin: int, end: int, node: LatticeNode) -> None:
        self.end_lists[end].append(node)
        node.begin = begin
        node.end = end
        self.connect_node(node)

    def remove(self, begin: int, end: int, node: LatticeNode) -> None:
        self.end_lists[end].remove(node)

    @staticmethod
    def create_node() -> LatticeNode:
        return LatticeNode()

    def has_previous_node(self, index: int) -> bool:
        return len(self.end_lists[index]) != 0

    def connect_node(self, r_node: LatticeNode) -> None:
        begin = r_node.begin
        r_node.total_cost = float('inf')
        for l_node in self.end_lists[begin]:
            if not l_node.is_connected_to_bos:
                continue
            # right_id and left_id look reversed, but it works ...
            connect_cost = self.grammar.get_connect_cost(l_node.right_id, r_node.left_id)
            if connect_cost == Grammar.INHIBITED_CONNECTION:
                continue
            cost = l_node.total_cost + connect_cost
            if cost < r_node.total_cost:
                r_node.total_cost = cost
                r_node.best_previous_node = l_node

        r_node.is_connected_to_bos = r_node.best_previous_node is not None
        r_node.total_cost += r_node.cost

    def connect_eos_node(self) -> None:
        self.connect_node(self.eos_node)

    def get_best_path(self) -> List[LatticeNode]:
        # self.connect_node(self.eos_node)
        if not self.eos_node.is_connected_to_bos:    # EOS node
            raise AttributeError("EOS is not connected to BOS")
        result = []
        node = self.eos_node
        while node is not self.end_lists[0][0]:
            result.append(node)
            node = node.best_previous_node
        return list(reversed(result))

    def dump(self, output):
        index = 0
        for i in range(self.size + 1, -1, -1):
            r_nodes = self.end_lists[i] if i <= self.size else [self.eos_node]
            for r_node in r_nodes:
                print("{}: {}: ".format(index, r_node), end="")
                index += 1
                for l_node in self.end_lists[r_node.begin]:
                    cost = l_node.total_cost + \
                        self.grammar.get_connect_cost(l_node.right_id, r_node.left_id)
                    print("{} ".format(cost), file=output, end="")
                print(file=output)



from sudachipy.dictionarylib.categorytype import CategoryType
from sudachipy.plugin.path_rewrite.path_rewrite_plugin import PathRewritePlugin


class JoinKatakanaOovPlugin(PathRewritePlugin):
    def __init__(self, json_obj):
        self.__pos = json_obj['oovPOS']
        self.oov_pos_id = None

    def set_up(self, grammar):
        # pos = ["名詞", "普通名詞", "一般", "*", "*", "*"]
        if not self.__pos:
            raise AttributeError("oovPOS is undefined")
        self.oov_pos_id = grammar.get_part_of_speech_id(self.__pos)
        if self.oov_pos_id < 0:
            raise AttributeError("oovPOS is invalid")

    def rewrite(self, text, path, lattice):
        i = 0
        while True:
            if i >= len(path):
                break
            node = path[i]
            if (node.is_oov or (self.is_one_char(text, node) and self.can_oov_bow_node(text, node))) and \
                    self.is_katakana_node(text, node):
                begin = i - 1
                while True:
                    if begin < 0:
                        break
                    if not self.is_katakana_node(text, path[begin]):
                        begin += 1
                        break
                    begin -= 1
                if begin < 0:
                    begin = 0
                while begin != i and not self.can_oov_bow_node(text, path[begin]):
                    begin += 1
                end = i + 1
                while True:
                    if end >= len(path):
                        break
                    if not self.is_katakana_node(text, path[end]):
                        break
                    end += 1
                if (end - begin) > 1:
                    self.concatenate_oov(path, begin, end, self.oov_pos_id, lattice)
                    i = begin + 1  # skip next node, as we already know it is not a joinable katakana
            i += 1

    def is_katakana_node(self, text, node):
        return CategoryType.KATAKANA in self.get_char_category_types(text, node)

    def is_one_char(self, text, node):
        b = node.get_begin()
        return b + text.get_code_points_offset_length(b, 1) == node.get_end()

    def can_oov_bow_node(self, text, node):
        return CategoryType.NOOOVBOW not in text.get_char_category_types(node.get_begin())



from abc import ABC, abstractmethod

from sudachipy.utf8inputtextbuilder import UTF8InputTextBuilder


class InputTextPlugin(ABC):

    Builder = UTF8InputTextBuilder

    @abstractmethod
    def set_up(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def rewrite(self, builder: Builder) -> None:
        raise NotImplementedError



from . import config
from . import dictionarylib
from .dictionarylib.binarydictionary import BinaryDictionary
from .dictionarylib.lexiconset import LexiconSet
from .plugin.input_text import DefaultInputTextPlugin
from .plugin.oov import get_oov_plugins
from .plugin.path_rewrite import get_path_rewrite_plugins
from .tokenizer import Tokenizer


class Dictionary:

    def __init__(self, config_path=None, resource_dir=None):
        config.settings.set_up(config_path, resource_dir)
        self.grammar = None
        self.lexicon = None
        self.input_text_plugins = []
        self.edit_connection_plugin = []
        self.oov_provider_plugins = []
        self.path_rewrite_plugins = []
        self.dictionaries = []
        self.header = None
        self._read_system_dictionary(config.settings.system_dict_path())

        # self.edit_connection_plugin = [InhibitConnectionPlugin()]
        # for p in self.edit_connection_plugin:
        #     p.set_up(self.grammar)
        #     p.edit(self.grammar)

        self._read_character_definition(config.settings.char_def_path())

        default_input_text_plugin = DefaultInputTextPlugin()
        self.input_text_plugins = [default_input_text_plugin]
        for p in self.input_text_plugins:
            p.set_up()

        self.oov_provider_plugins = get_oov_plugins()
        if not self.oov_provider_plugins:
            raise AttributeError("no OOV provider")
        for p in self.oov_provider_plugins:
            p.set_up(self.grammar)

        self.path_rewrite_plugins = get_path_rewrite_plugins()
        for p in self.path_rewrite_plugins:
            p.set_up(self.grammar)

        for filename in config.settings.user_dict_paths():
            self._read_user_dictionary(filename)

    def _read_system_dictionary(self, filename):
        if filename is None:
            raise AttributeError("system dictionary is not specified")
        dict_ = BinaryDictionary.from_system_dictionary(filename)
        self.dictionaries.append(dict_)
        self.grammar = dict_.grammar
        self.lexicon = LexiconSet(dict_.lexicon)

    def _read_user_dictionary(self, filename):
        if self.lexicon.is_full():
            raise ValueError('too many dictionaries')
        dict_ = BinaryDictionary.from_user_dictionary(filename)
        self.dictionaries.append(dict_)
        user_lexicon = dict_.lexicon
        tokenizer_ = Tokenizer(self.grammar, self.lexicon, self.input_text_plugins, self.oov_provider_plugins, [])
        user_lexicon.calculate_cost(tokenizer_)
        self.lexicon.add(user_lexicon, self.grammar.get_part_of_speech_size())
        if dict_.grammar:
            self.grammar.add_pos_list(dict_.grammar)

    def _read_character_definition(self, filename):
        if self.grammar is None:
            return
        char_category = dictionarylib.charactercategory.CharacterCategory()
        char_category.read_character_definition(filename)
        self.grammar.set_character_category(char_category)

    def close(self):
        self.grammar = None
        self.lexicon = None
        for dict_ in self.dictionaries:
            dict_.close()

    def create(self):
        return Tokenizer(
            self.grammar, self.lexicon, self.input_text_plugins, self.oov_provider_plugins, self.path_rewrite_plugins)



import os
from unicodedata import normalize

from sudachipy import config

from . import InputTextPlugin


class DefaultInputTextPlugin(InputTextPlugin):
    def __init__(self):
        self.ignore_normalize_set = set()
        self.key_lengths = {}
        self.replace_char_map = {}

    def set_up(self) -> None:
        rewrite_def = os.path.join(config.DEFAULT_RESOURCEDIR, "rewrite.def")
        if not rewrite_def:
            raise AttributeError("rewriteDef is not defined")
        self.read_rewrite_lists(rewrite_def)

    def rewrite(self, builder: InputTextPlugin.Builder) -> None:
        offset = 0
        next_offset = 0
        text = builder.get_text()

        i = -1
        while True:
            i += 1
            if i >= len(text):
                break
            textloop = False
            offset += next_offset
            next_offset = 0
            original = text[i]

            # 1. replace char without normalize
            max_length = min(self.key_lengths.get(original, 0), len(text) - i)
            for l in range(max_length, 0, -1):
                replace = self.replace_char_map.get(text[i:i + l])
                if replace:
                    builder.replace(i + offset, i + l + offset, replace)
                    next_offset += len(replace) - l
                    i += l - 1
                    textloop = True
                    break
            if textloop:
                continue

            # 2. normalize
            # 2-1. capital alphabet (not only Latin but Greek, Cyrillic, etc.) -> small
            lower = original.lower()
            if lower in self.ignore_normalize_set:
                if original == lower:
                    continue
                replace = lower
            else:
                # 2-2. normalize (except in ignoreNormalize)
                #   e.g. full-width alphabet -> half-width / ligature / etc.
                replace = normalize("NFKC", lower)
            next_offset = len(replace) - 1
            if original != replace:
                builder.replace(i + offset, i + 1 + offset, replace)

    def read_rewrite_lists(self, rewrite_def):
        with open(rewrite_def, "r", encoding="utf-8") as f:
            for i, line in enumerate(f):
                line = line.strip()
                if (not line) or line.startswith("#"):
                    continue
                cols = line.split()

                # ignored normalize list
                if len(cols) == 1:
                    key = cols[0]
                    if len(key) != 1:
                        raise RuntimeError("{} is not character at line {}".format(key, i))
                    self.ignore_normalize_set.add(key)
                # replace char list
                elif len(cols) == 2:
                    if cols[0] in self.replace_char_map:
                        raise RuntimeError("{} is already defined at line {}".format(cols[0], i))
                    if self.key_lengths.get(cols[0][0], 0) < len(cols[0]):
                        self.key_lengths[cols[0][0]] = len(cols[0])
                    self.replace_char_map[cols[0]] = cols[1]
                else:
                    raise RuntimeError("invalid format at line {}".format(i))



import struct


class WordIdTable(object):
    def __init__(self, bytes_, offset):
        self.bytes = bytes_
        bytes_.seek(offset)
        self.size = int.from_bytes(bytes_.read(4), 'little')
        self.offset = offset + 4

    def storage_size(self):
        return 4 + self.size

    def get(self, index):
        length = self.bytes[self.offset + index]
        index += 1
        result = struct.unpack_from("<{}I".format(length), self.bytes, self.offset + index)
        return result



class WordInfo:
    def __init__(self,
                 surface,
                 head_word_length,
                 pos_id,
                 normalized_form,
                 dictionary_form_word_id,
                 dictionary_form,
                 reading_form,
                 a_unit_split,
                 b_unit_split,
                 word_structure):
        self.surface = surface
        self.head_word_length = head_word_length
        self.pos_id = pos_id
        self.normalized_form = normalized_form
        self.dictionary_form_word_id = dictionary_form_word_id
        self.dictionary_form = dictionary_form
        self.reading_form = reading_form
        self.a_unit_split = a_unit_split
        self.b_unit_split = b_unit_split
        self.word_structure = word_structure

    def length(self):
        return self.head_word_length



import mmap

from . import doublearraybuilder
from . import keyset


class DoubleArray(object):

    def __init__(self):
        self.array = None
        self.buffer = None
        self.size = 0

    def set_array(self, array, size):
        """
        :param array: bytes(int)
        """
        self.array = array
        self.size = size

    def array(self):
        return self.array

    def byte_array(self):
        return self.buffer

    def clear(self):
        self.buffer = None
        self.size = 0

    # def size(self):
    #     return self.size

    def total_size(self):
        return 4 * self.size

    def build(self, keys, values, progress_function):
        key_set = keyset.KeySet(keys, values)
        builder = doublearraybuilder.DoubleArrayBuilder(progress_function)
        builder.build(key_set)
        self.buffer = builder.copy()
        bytes_ = self.buffer.getvalue()
        self.array = [
            int.from_bytes(bytes_[4 * i: 4 * i + 4], byteorder='little')
            for i in range(len(bytes_) // 4)]
        self.size = len(self.array)
        # for i, binary in enumerate(self.array):
        #     print("{0:03d}".format(i), "{0:032b}".format(binary), file=sys.stderr)

    def open(self, input_file, position, total_size):
        if position < 0:
            position = 0

        self.buffer = mmap.mmap(input_file.fileno(), total_size, offset=position)
        self.array = self.buffer.as_int_buffer()
        self.size = self.array.capacity()

    def save(self, output_file):
        pass

    def exact_match_search(self, key):
        result = [-1, 0]
        node_pos = 0
        unit = self.array[node_pos]

        for k in key:
            node_pos ^= self.offset(unit) ^ k
            unit = self.array[node_pos]
            if (self.label(unit) is not k):
                return result
        if not self.has_leaf(unit):
            return result
        unit = self.array[node_pos ^ self.offset(unit)]
        result[0] = self.value(unit)
        result[1] = len(key)
        return result

    def common_prefix_search(self, key, offset):
        result = []

        node_pos = 0
        unit = self.array[node_pos]
        node_pos ^= self.offset(unit)

        for i in range(offset, len(key)):
            k = key[i]
            node_pos ^= k
            unit = self.array[node_pos]
            if self.label(unit) != k:
                return result

            node_pos ^= self.offset(unit)
            if (self.has_leaf(unit)):
                r = [self.value(self.array[node_pos]), i + 1]
                result.append(r)
        return result

    def has_leaf(self, unit):
        return ((unit >> 8) & 1) == 1

    def value(self, unit):
        return unit & ((1 << 31) - 1)

    def label(self, unit):
        return unit & ((1 << 31) | 0xFF)

    def offset(self, unit):
        return (unit >> 10) << ((unit & (1 << 9)) >> 6)

    def byte_to_u_int(self, b):
        return int.from_bytes(b, 'little', signed=False)



from io import BytesIO


class JTypedByteBuffer(BytesIO):
    """
    A interface of BytesIO to write dictionary
    """

    __ENDIAN = 'little'

    @classmethod
    def from_bytes(cls, bytes_io):
        return cls(bytes_io.getvalue())

    def write_int(self, int_, type_, signed=True):
        if type_ == 'byte':
            len_ = 1
            signed = False
        elif type_ == 'int':
            len_ = 4
        elif type_ == 'char':
            len_ = 2
            signed = False
        elif type_ == 'short':
            len_ = 2
        elif type_ == 'long':
            len_ = 8
        else:
            raise ValueError('{} is invalid type'.format(type_))
        self.write(int_.to_bytes(len_, byteorder=self.__ENDIAN, signed=signed))

    def write_str(self, text):
        self.write(text.encode('utf-16-le'))

    def clear(self):
        self.seek(0)
        self.truncate(0)



from enum import Enum
from typing import List

from .dictionarylib.categorytype import CategoryType
from .dictionarylib.grammar import Grammar
from .dictionarylib.lexicon import Lexicon
from .lattice import Lattice
from .latticenode import LatticeNode
from .morphemelist import MorphemeList
from .plugin.input_text import InputTextPlugin
from .plugin.path_rewrite import PathRewritePlugin
from .utf8inputtext import UTF8InputText
from .utf8inputtextbuilder import UTF8InputTextBuilder


class Tokenizer:
    SplitMode = Enum("SplitMode", "A B C")

    def __init__(self, grammar: Grammar, lexicon: Lexicon, input_text_plugins: List[InputTextPlugin],
                 oov_provider_plugins: List, path_rewrite_plugins: List[PathRewritePlugin]):
        self.grammar = grammar
        self.lexicon = lexicon
        self.input_text_plugins = input_text_plugins
        self.oov_provider_plugins = oov_provider_plugins
        self.path_rewrite_plugins = path_rewrite_plugins
        self.dump_output = None
        self.lattice = Lattice(grammar)

        if self.oov_provider_plugins:
            self.default_oov_provider = self.oov_provider_plugins[-1]

    def build_lattice(self, input_: UTF8InputText):
        bytes_ = input_.get_byte_text()
        self.lattice.resize(len(bytes_))
        for i in range(len(bytes_)):
            if not input_.is_char_alignment(i) or not self.lattice.has_previous_node(i):
                continue
            iterator = self.lexicon.lookup(bytes_, i)
            has_words = False
            for word_id, end in iterator:
                has_words = True
                n = LatticeNode(self.lexicon,
                                self.lexicon.get_left_id(word_id),
                                self.lexicon.get_right_id(word_id),
                                self.lexicon.get_cost(word_id),
                                word_id)
                self.lattice.insert(i, end, n)

            # OOV
            if CategoryType.NOOOVBOW not in input_.get_char_category_types(i):
                for oov_plugin in self.oov_provider_plugins:
                    for node in oov_plugin.get_oov(input_, i, has_words):
                        has_words = True
                        self.lattice.insert(node.get_begin(), node.get_end(), node)
            if not has_words and self.default_oov_provider:
                for node in self.default_oov_provider.get_oov(input_, i, has_words):
                    has_words = True
                    self.lattice.insert(node.get_begin(), node.get_end(), node)

            if not has_words:
                raise AttributeError("there is no morpheme at " + str(i))
        self.lattice.connect_eos_node()

    def split_path(self, path: List[LatticeNode], mode: SplitMode) -> List[LatticeNode]:
        new_path = []
        for node in path:
            if mode is self.SplitMode.A:
                wids = node.get_word_info().a_unit_split
            else:  # self.SplitMode.B
                wids = node.get_word_info().b_unit_split
            if 0 <= len(wids) <= 1:
                new_path.append(node)
            else:
                offset = node.get_begin()
                for wid in wids:
                    n = LatticeNode(self.lexicon, 0, 0, 0, wid)
                    n.begin = offset
                    offset += n.get_word_info().head_word_length
                    n.end = offset
                    new_path.append(n)
        return new_path

    def tokenize(self, text: str, mode=None) -> MorphemeList:
        if not text:
            return []
        mode = mode or self.SplitMode.C

        builder = UTF8InputTextBuilder(text, self.grammar)
        for plugin in self.input_text_plugins:
            plugin.rewrite(builder)
        input_ = builder.build()
        # dump
        self.build_lattice(input_)

        if self.dump_output:
            print("=== Lattice dump:", file=self.dump_output)
            self.lattice.dump(self.dump_output)
        path = self.lattice.get_best_path()
        # dump
        path.pop()  # remove EOS
        # dump_output
        for plugin in self.path_rewrite_plugins:
            plugin.rewrite(input_, path, self.lattice)
        self.lattice.clear()
        if mode is not self.SplitMode.C:
            path = self.split_path(path, mode)
        # dump_output
        ml = MorphemeList(input_, self.grammar, self.lexicon, path)
        return ml

    def set_dump_output(self, output):
        self.dump_output = output



from abc import ABC, abstractmethod

from sudachipy.dictionarylib.wordinfo import WordInfo


class PathRewritePlugin(ABC):

    @abstractmethod
    def set_up(self, grammar):
        raise NotImplementedError

    @abstractmethod
    def rewrite(self, text, path, lattice):
        raise NotImplementedError

    def concatenate(self, path, begin, end, lattice):
        if begin >= end:
            raise IndexError("begin >= end")
        b = path[begin].get_begin()
        e = path[end - 1].get_end()
        pos_id = path[begin].get_word_info().pos_id
        surface = ""
        length = 0
        normalized_form, dictionary_form, reading_form = "", "", ""
        for i in range(begin, end):
            info = path[i].get_word_info()
            surface += info.surface
            length += info.head_word_length
            normalized_form += info.normalized_form
            dictionary_form += info.dictionary_form
            reading_form += info.reading_form

        wi = WordInfo(surface=surface, head_word_length=length, pos_id=pos_id,
                      normalized_form=normalized_form, dictionary_form=dictionary_form, dictionary_form_word_id=-1,
                      reading_form=reading_form, a_unit_split=[], b_unit_split=[], word_structure=[])

        node = lattice.create_node()
        node.set_range(b, e)
        node.set_word_info(wi)

        path[begin:end] = [node]
        return node

    def concatenate_oov(self, path, begin, end, pos_id, lattice):
        if begin >= end:
            raise IndexError("begin >= end")
        b = path[begin].get_begin()
        e = path[end - 1].get_end()
        surface = ""
        length = 0
        for i in range(begin, end):
            info = path[i].get_word_info()
            surface += info.surface
            length += info.head_word_length

        wi = WordInfo(surface=surface, head_word_length=length, pos_id=pos_id,
                      normalized_form=surface, dictionary_form=surface, dictionary_form_word_id=-1,
                      reading_form="", a_unit_split=[], b_unit_split=[], word_structure=[])

        node = lattice.create_node()
        node.set_range(b, e)
        node.set_word_info(wi)
        node.set_oov()

        path[begin:end] = [node]
        return node

    def get_char_category_types(self, text, node):
        return text.get_char_category_types(node.get_begin(), node.get_end())



import json
import os
from typing import List

DEFAULT_SETTINGFILE = os.path.join(os.path.dirname(os.path.abspath(__file__)), os.pardir, "resources/sudachi.json")
DEFAULT_RESOURCEDIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), os.pardir, "resources")


class _Settings(object):

    def __init__(self):
        self.__is_active = False
        self.__dict_ = None
        self.resource_dir = DEFAULT_RESOURCEDIR

    def set_up(self, path=None, resource_dir=None) -> None:
        if not path:
            path = DEFAULT_SETTINGFILE
        if not resource_dir:
            resource_dir = DEFAULT_RESOURCEDIR
        with open(path, "r", encoding="utf-8") as f:
            self.__dict_ = json.load(f)
        self.__is_active = True
        self.resource_dir = resource_dir

    def __getitem__(self, key):
        if not self.__is_active:
            self.set_up()
        return self.__dict_[key]

    def keys(self):
        return self.__dict_.keys()

    def __contains__(self, item):
        return item in self.__dict_.keys()

    def system_dict_path(self) -> str:
        if 'systemDict' in self.__dict_:
            return os.path.join(self.resource_dir, self.__dict_['systemDict'])
        raise KeyError('`systemDict` not defined in setting file')

    def char_def_path(self) -> str:
        if 'characterDefinitionFile' in self.__dict_:
            return os.path.join(self.resource_dir, self.__dict_['characterDefinitionFile'])
        raise KeyError('`characterDefinitionFile` not defined in setting file')

    def user_dict_paths(self) -> List[str]:
        if 'userDict' in self.__dict_:
            return [os.path.join(self.resource_dir, path) for path in self.__dict_['userDict']]
        return []


settings = _Settings()



import copy

from . import utf8inputtext


class UTF8InputTextBuilder:
    def __init__(self, text, grammar):

        self.grammar = grammar
        self.original_text = text
        self.modified_text = text
        self.text_offsets = list(range(len(self.original_text) + 1))
        # 注: サロゲートペア文字は考慮していない

    def replace(self, begin, end, str_):
        if begin < 0:
            raise IndexError(begin)
        if begin > len(self.modified_text):
            raise IndexError("begin > length")
        if begin > end:
            raise IndexError("begin > end")
        if begin == end:
            raise AttributeError("begin == end")

        if end > len(self.modified_text):
            end = len(self.modified_text)

        self.modified_text = str_.join([self.modified_text[:begin], self.modified_text[end:]])

        offset = self.text_offsets[begin]
        length = len(str_)
        if end - begin > length:
            del self.text_offsets[begin + length:end]
        for i in range(length):
            if begin + i < end:
                self.text_offsets[begin + i] = offset
            else:
                self.text_offsets.insert(begin + i, offset)

    def get_original_text(self):
        return self.original_text

    def get_text(self):
        return self.modified_text

    def build(self):
        modified_string_text = self.get_text()
        byte_text = modified_string_text.encode('utf-8')

        length = len(byte_text)
        byte_indexes = [0 for i in range(length + 1)]
        offsets = [0 for i in range(length + 1)]
        j = 0
        for i in range(len(self.modified_text)):
            # 注: サロゲートペア文字は考慮していない
            for _ in range(self.utf8_byte_length(ord(self.modified_text[i]))):
                byte_indexes[j] = i
                offsets[j] = self.text_offsets[i]
                j += 1
        byte_indexes[length] = len(modified_string_text)
        offsets[length] = self.text_offsets[-1]

        char_categories = self.get_char_category_types(modified_string_text)
        char_category_continuities = self.get_char_category_continuities(modified_string_text, length, char_categories)
        return utf8inputtext.UTF8InputText(self.grammar, self.original_text, modified_string_text, byte_text, offsets, byte_indexes, char_categories, char_category_continuities)

    def get_char_category_types(self, text):
        if len(text) == 0:
            return []

        char_category_types = [None for i in range(len(text))]
        for i in range(len(text)):
            types = self.grammar.get_character_category().get_category_types(ord(text[i]))
            char_category_types[i] = types
        return char_category_types

    def get_char_category_continuities(self, text, byte_length, char_categories):
        if len(text) == 0:
            return []
        char_category_continuities = []
        i = 0
        while i < len(char_categories):
            next_ = i + self.get_char_category_continuous_length(char_categories, i)
            length = 0
            for j in range(i, next_):
                length += self.utf8_byte_length(ord(text[j]))
            for k in range(length, 0, -1):
                char_category_continuities.append(k)
            i = next_
        return char_category_continuities

    def get_char_category_continuous_length(self, char_categories, offset):
        continuous_category = copy.deepcopy(char_categories[offset])
        for length in range(1, len(char_categories) - offset):
            continuous_category = continuous_category & char_categories[offset + length]
            if len(continuous_category) == 0:
                return length
        return len(char_categories) - offset

    def utf8_byte_length(self, cp):
        if cp < 0:
            return 0
        elif cp <= 0x7F:
            return 1
        elif cp <= 0x7FF:
            return 2
        elif cp <= 0xFFFF:
            return 3
        elif cp <= 0x10FFFF:
            return 4
        else:
            return 0



from setuptools import setup, find_packages

setup(name="SudachiPy",
      version="0.2.0.1",
      description="Python version of Sudachi, the Japanese Morphological Analyzer",
      long_description=open('README.md').read(),
      long_description_content_type="text/markdown",
      url="https://github.com/WorksApplications/SudachiPy",
      license="Apache-2.0",
      author="Works Applications",
      author_email="takaoka_k@worksap.co.jp",
      packages=find_packages(include=["sudachipy", "sudachipy.*"]),
      entry_points={
          "console_scripts": ["sudachipy=sudachipy.command_line:main"],
      },
      install_requires=["sortedcontainers>=2.1.0,<2.2.0"],
      include_package_data=True,
      )



class KeySet(object):
    def __init__(self, keys, values):
        self.keys = keys
        self.values = values

    def size(self):
        return len(self.keys)

    def get_key(self, id):
        return self.keys[id]

    def get_key_byte(self, key_id, byte_id):
        if byte_id >= len(self.keys[key_id]):
            return 0
        return self.keys[key_id][byte_id]

    def has_values(self):
        return self.values is not None

    def get_value(self, id):
        return self.values[id] if self.has_values() else id



import copy


class UTF8InputText:
    def __init__(self, grammar, original_text, modified_text, bytes_, offsets, byte_indexes, char_categories, char_category_continuities, can_bow_list=None):
        self.original_text = original_text
        self.modified_text = modified_text
        self.bytes = bytes_
        self.offsets = offsets
        self.byte_indexes = byte_indexes
        self.char_categories = char_categories
        self.char_category_continuities = char_category_continuities
        self.can_bow_list = can_bow_list

    def get_original_text(self):
        return self.original_text

    def get_text(self):
        return self.modified_text

    def get_byte_text(self):
        return self.bytes

    def get_substring(self, begin, end):
        if begin < 0:
            raise IndexError(begin)
        if end > len(self.bytes):
            raise IndexError(end)
        if (begin > end):
            raise IndexError(end - begin)

        return self.modified_text[self.byte_indexes[begin]:self.byte_indexes[end]]

    def get_offset_text_length(self, index):
        return self.byte_indexes[index]

    def is_char_alignment(self, index):
        return (self.bytes[index] & 0xC0) is not 0x80

    def get_original_index(self, index):
        return self.offsets[index]

    def get_char_category_types(self, begin, end=None):
        if end is None:
            return self.char_categories[self.byte_indexes[begin]]
        if begin + self.get_char_category_continuous_length(begin) < end:
            return []
        b = self.byte_indexes[begin]
        e = self.byte_indexes[end]
        continuous_category = copy.deepcopy(self.char_categories[b])
        for i in range(b + 1, e):
            continuous_category = continuous_category & self.char_categories[i]
        return continuous_category

    def get_char_category_continuous_length(self, index):
        return self.char_category_continuities[index]

    def get_code_points_offset_length(self, index, code_point_offset):
        length = 0
        target = self.byte_indexes[index] + code_point_offset
        for i in range(index, len(self.bytes)):
            if self.byte_indexes[i] >= target:
                return length
            length += 1
        return length

    def can_bow(self, idx: int) -> bool:
        return self.is_char_alignment(idx) and self.can_bow_list[self.byte_indexes[idx]]



from . import wordinfo


class WordInfoList(object):
    def __init__(self, bytes_, offset, word_size):
        self.bytes = bytes_
        self.offset = offset
        self._word_size = word_size

    def get_word_info(self, word_id):
        orig_pos = self.bytes.tell()
        index = self.word_id_to_offset(word_id)
        self.bytes.seek(index)
        surface = self.buffer_to_string()
        head_word_length = self.buffer_to_string_length()
        pos_id = int.from_bytes(self.bytes.read(2), 'little')
        normalized_form = self.buffer_to_string()
        if not normalized_form:
            normalized_form = surface
        dictionary_form_word_id = int.from_bytes(self.bytes.read(4), 'little', signed=True)
        reading_form = self.buffer_to_string()
        a_unit_split = self.buffer_to_int_array()
        b_unit_split = self.buffer_to_int_array()
        word_structure = self.buffer_to_int_array()

        dictionary_form = surface
        if dictionary_form_word_id >= 0 and dictionary_form_word_id != word_id:
            wi = self.get_word_info(dictionary_form_word_id)
            dictionary_form = wi.surface

        self.bytes.seek(orig_pos)

        return wordinfo.WordInfo(surface, head_word_length, pos_id, normalized_form,
                                 dictionary_form_word_id, dictionary_form, reading_form,
                                 a_unit_split, b_unit_split, word_structure)

    def word_id_to_offset(self, word_id):
        i = self.offset + 4 * word_id
        return int.from_bytes(self.bytes[i:i + 4], 'little', signed=False)

    def buffer_to_string_length(self):
        length = self.bytes.read_byte()
        if length < 128:
            return length
        low = self.bytes.read_byte()
        return ((length & 0x7F) << 8) | low

    def buffer_to_string(self):
        length = self.buffer_to_string_length()
        return self.bytes.read(2 * length).decode('utf-16-le')

    def buffer_to_int_array(self):
        length = self.bytes.read_byte()
        array = []
        for _ in range(length):
            array.append(int.from_bytes(self.bytes.read(4), 'little', signed=True))
        return array

    def size(self):
        return self._word_size



from io import BytesIO

from . import dawgbuilder
from . import doublearraybuilderunit


class DoubleArrayBuilder(object):

    class DoubleArrayBuilderExtraUnit(object):
        def __init__(self):
            self.prev = 0
            self.next = 0
            self.is_fixed = False
            self.is_used = False

    def __init__(self, progress_function):
        self.BLOCK_SIZE = 256
        self.NUM_EXTRA_BLOCKS = 16
        self.NUM_EXTRAS = self.BLOCK_SIZE * self.NUM_EXTRA_BLOCKS
        self.UPPER_MASK = 0xFF << 21
        self.LOWER_MASK = 0xFF

        self.progress_function = progress_function
        self.clear()
        self.extras_head = 0

    def build(self, key_set):
        if key_set.has_values():
            dawg_builder = dawgbuilder.DAWGBuilder()
            self.build_dawg(key_set, dawg_builder)
            self.build_from_dawg_header(dawg_builder)
            dawg_builder.clear()
        else:
            self.build_from_key_set_header(key_set)

    def copy(self):
        buf = BytesIO()
        for u in self.units:
            buf.write(u.unit.to_bytes(4, byteorder='little', signed=False))
            # buf.write_int(u.unit, 'int', signed=False)
        buf.seek(0)
        return buf

    def clear(self):
        self.units = []
        self.extras = []
        self.labels = []
        self.table = []

    def num_blocks(self):
        return int(len(self.units) / self.BLOCK_SIZE)

    def get_extras(self, id):
        return self.extras[id % self.NUM_EXTRAS]

    def build_dawg(self, key_set, dawg_builder):
        dawg_builder.init()
        for i in range(key_set.size()):
            dawg_builder.insert(key_set.get_key(i), key_set.get_value(i))
            if self.progress_function is not None:
                # self.progress_function.accept(i + 1, key_set.size() + 1)
                self.progress_function(i + 1, key_set.size() + 1)
        dawg_builder.finish()

    def build_from_dawg_header(self, dawg):
        """
        num_units = 1
        while num_units < dawg.size():
            num_units <<= 1
        """

        self.table = [0] * dawg.num_intersections()
        self.extras = [self.DoubleArrayBuilderExtraUnit() for i in range(self.NUM_EXTRAS)]

        self.reserve_id(0)
        self.get_extras(0).is_used = True
        self.units[0].set_offset(1)
        self.units[0].set_label(0)

        if dawg.child(dawg.root()) is not 0:
            self.build_from_dawg_insert(dawg, dawg.root(), 0)

        self.fix_all_blocks()

        self.extras = None
        self.labels = None
        self.table = None

    def build_from_dawg_insert(self, dawg, dawg_id, dic_id):
        dawg_child_id = dawg.child(dawg_id)
        if dawg.is_intersection(dawg_child_id):
            intersection_id = dawg.intersection_id(dawg_child_id)
            offset = self.table[intersection_id]
            if offset is not 0:
                offset ^= dic_id
                if (offset & self.UPPER_MASK) is 0 or (offset & self.LOWER_MASK) is 0:
                    if dawg.is_leaf(dawg_child_id):
                        self.units[dic_id].set_has_leaf(True)
                    self.units[dic_id].set_offset(offset)
                    return

        offset = self.arrange_from_dawg(dawg, dawg_id, dic_id)
        if dawg.is_intersection(dawg_child_id):
            self.table[dawg.intersection_id(dawg_child_id)] = offset

        while True:
            child_label = dawg.label(dawg_child_id)
            dic_child_id = offset ^ child_label
            if child_label is not 0:
                self.build_from_dawg_insert(dawg, dawg_child_id, dic_child_id)
            dawg_child_id = dawg.sibling(dawg_child_id)
            if dawg_child_id is 0:
                break

    def arrange_from_dawg(self, dawg, dawg_id, dic_id):
        self.labels.clear()

        dawg_child_id = dawg.child(dawg_id)
        while(dawg_child_id is not 0):
            self.labels.append(dawg.label(dawg_child_id))
            dawg_child_id = dawg.sibling(dawg_child_id)

        offset = self.find_valid_offset(dic_id)
        self.units[dic_id].set_offset(dic_id ^ offset)

        dawg_child_id = dawg.child(dawg_id)
        for l in self.labels:
            dic_child_id = offset ^ l
            self.reserve_id(dic_child_id)

            if dawg.is_leaf(dawg_child_id):
                self.units[dic_id].set_has_leaf(True)
                self.units[dic_child_id].set_value(dawg.value(dawg_child_id))
            else:
                self.units[dic_child_id].set_label(l)

            dawg_child_id = dawg.sibling(dawg_child_id)
        self.get_extras(offset).is_used = True
        return offset

    def build_from_key_set_header(self, key_set):
        """
        num_units = 1
        while num_units < key_set.size():
            num_units <<= 1
        """

        self.extras = [self.DoubleArrayBuilderExtraUnit() for i in range(self.NUM_EXTRAS)]
        self.extras[2].next = 1

        self.reserve_id(0)
        self.get_extras(0).is_used = True
        self.units[0].set_offset(1)
        self.units[0].set_label(0)

        if key_set.size() > 0:
            self.build_from_key_set_insert(key_set, 0, key_set.size(), 0, 0)

        self.fix_all_blocks()

        self.extras = None
        self.labels = None

    def build_from_key_set_insert(self, key_set, begin, end, depth, dic_id):
        """
        :param key_set: KeySet
        :param begin: int
        :param end: int
        :param depth: int
        :param dic_id: int
        """
        offset = self.arrange_from_key_set(key_set, begin, end, depth, dic_id)

        while begin < end:
            if key_set.get_key_byte(begin, depth) is not 0:
                break
            begin += 1
        if begin is end:
            return

        last_begin = begin
        last_label = key_set.get_key_byte(begin, depth)
        while True:
            begin += 1
            if begin >= end:
                break
            label = key_set.get_key_byte(begin, depth)
            if label is not last_label:
                self.build_from_key_set_insert(key_set, last_begin, begin, depth + 1, offset ^ last_label)
                last_begin = begin
                last_label = key_set.get_key_byte(begin, depth)
        self.build_from_key_set_insert(key_set, last_begin, end, depth + 1, offset ^ last_label)

    def arrange_from_key_set(self, key_set, begin, end, depth, dic_id):

        self.labels.clear()

        value = -1
        for i in range(begin, end):
            label = key_set.get_key_byte(i, depth)
            if label is 0:
                if depth < len(key_set.get_key(i)):
                    raise AttributeError("invalid null character")
                elif key_set.get_value(i) < 0:
                    raise AttributeError("negative value")

                if value is -1:
                    value = key_set.get_value(i)
                if self.progress_function is not None:
                    # self.progress_function.accept(i + 1, key_set.size() + 1)
                    self.progress_function(i + 1, key_set.size() + 1)

            if len(self.labels) is 0:
                self.labels.append(label)
            elif label is not self.labels[-1]:
                if label < self.labels[-1]:
                    raise AttributeError("wrong key order")
                self.labels.append(label)

        offset = self.find_valid_offset(dic_id)
        self.units[dic_id].set_offset(dic_id ^ offset)

        for l in self.labels:
            dic_child_id = offset ^ l
            self.reserve_id(dic_child_id)
            if (l is 0):
                self.units[dic_id].set_has_leaf(True)
                self.units[dic_child_id].set_value(value)
            else:
                self.units[dic_child_id].set_label(l)
        self.get_extras(offset).is_used = True

        return offset

    def find_valid_offset(self, id):
        if self.extras_head >= len(self.units):
            return len(self.units) | (id & self.LOWER_MASK)

        unfixed_id = self.extras_head
        memo = []
        while True:
            # debug
            if unfixed_id not in memo:
                memo.append(unfixed_id)
            else:
                raise RuntimeError(unfixed_id, memo)

            offset = unfixed_id ^ self.labels[0]
            if self.is_valid_offset(id, offset):
                return offset
            unfixed_id = self.get_extras(unfixed_id).next

            if unfixed_id is self.extras_head:
                break

        return len(self.units) | (id & self.LOWER_MASK)

    def is_valid_offset(self, id, offset):
        if self.get_extras(offset).is_used:
            return False

        rel_offset = id ^ offset
        if (rel_offset & self.LOWER_MASK) is not 0 and (rel_offset & self.UPPER_MASK) is not 0:
            return False

        for i in range(1, len(self.labels)):
            if self.get_extras(offset ^ self.labels[i]).is_fixed:
                return False

        return True

    def reserve_id(self, id):
        if id >= len(self.units):
            self.expand_units()

        if id is self.extras_head:
            self.extras_head = self.get_extras(id).next
            if self.extras_head is id:
                self.extras_head = len(self.units)
        self.get_extras(self.get_extras(id).prev).next = self.get_extras(id).next
        self.get_extras(self.get_extras(id).next).prev = self.get_extras(id).prev
        self.get_extras(id).is_fixed = True

    def expand_units(self):
        src_num_units = len(self.units)
        src_num_blocks = self.num_blocks()

        dest_num_units = src_num_units + self.BLOCK_SIZE
        dest_num_blocks = src_num_blocks + 1

        if dest_num_blocks > self.NUM_EXTRA_BLOCKS:
            self.fix_block(src_num_blocks - self.NUM_EXTRA_BLOCKS)

        for i in range(src_num_units, dest_num_units):
            self.units.append(doublearraybuilderunit.DoubleArrayBuilderUnit())

        if dest_num_blocks > self.NUM_EXTRA_BLOCKS:
            for id_ in range(src_num_units, dest_num_units):
                self.get_extras(id_).is_used = False
                self.get_extras(id_).is_fixed = False

        for i in range(src_num_units + 1, dest_num_units):
            self.get_extras(i - 1).next = i
            self.get_extras(i).prev = i - 1

        self.get_extras(src_num_units).prev = dest_num_units - 1
        self.get_extras(dest_num_units - 1).next = src_num_units
        # It seems to me that meaningless because reassign in following
        # 2017.08.24 Hiyowa Kyobashi

        self.get_extras(src_num_units).prev = self.get_extras(self.extras_head).prev
        self.get_extras(dest_num_units - 1).next = self.extras_head

        self.get_extras(self.get_extras(self.extras_head).prev).next = src_num_units
        self.get_extras(self.extras_head).prev = dest_num_units - 1

    def fix_all_blocks(self):
        begin = 0
        if self.num_blocks() > self.NUM_EXTRA_BLOCKS:
            begin = self.num_blocks() - self.NUM_EXTRA_BLOCKS
        end = self.num_blocks()

        for block_id in range(begin, end):
            self.fix_block(block_id)

    def fix_block(self, block_id):
        begin = block_id * self.BLOCK_SIZE
        end = begin + self.BLOCK_SIZE

        unused_offset = 0
        offset = begin

        while offset is not end:
            if not self.get_extras(offset).is_used:
                unused_offset = offset
                break
            offset += 1

        for id_ in range(begin, end):
            if not self.get_extras(id_).is_fixed:
                self.reserve_id(id_)
                self.units[id_].set_label(id_ ^ unused_offset)

    def byte_to_u_int(self, b):
        return int.from_bytes(b, 'little', signed=False)



from sudachipy.dictionarylib.categorytype import CategoryType

from .path_rewrite_plugin import PathRewritePlugin


class JoinNumericPlugin(PathRewritePlugin):
    def __init__(self, json_obj):
        self.join_kanji_numeric = True
        if 'enableNormalize' in json_obj:
            self.join_kanji_numeric = json_obj['enableNormalize']
        self.join_all_numeric = None

    def set_up(self, grammar):
        self.join_all_numeric = False

    def rewrite(self, text, path, lattice):
        begin_index = -1
        type_ = None

        i = -1
        while True:
            i += 1
            if i >= len(path):
                break
            node = path[i]
            types = self.get_char_category_types(text, node)
            if CategoryType.NUMERIC in types:
                if type_ == CategoryType.NUMERIC:
                    continue
                if type_ == CategoryType.KANJINUMERIC:
                    if self.join_all_numeric:
                        continue
                    elif self.join_kanji_numeric:
                        if (i - begin_index) > 1:
                            self.concatenate(path, begin_index, i, lattice)
                            i = begin_index + 1
                type_ = CategoryType.NUMERIC
                begin_index = i
            elif CategoryType.KANJINUMERIC in types:
                if type_ == CategoryType.KANJINUMERIC:
                    continue
                if type_ == CategoryType.NUMERIC:
                    if self.join_all_numeric:
                        continue
                    if (i - begin_index) > 1:
                        self.concatenate(path, begin_index, i, lattice)
                    i = begin_index + 1
                if self.join_kanji_numeric:
                    type_ = CategoryType.KANJINUMERIC
                    begin_index = i
                else:
                    type_ = None
                    begin_index = -1
            else:
                if begin_index >= 0:
                    if (i - begin_index) > 1:
                        self.concatenate(path, begin_index, i, lattice)
                    i = begin_index + 1
                type_ = None
                begin_index = -1
        if begin_index >= 0 and (len(path) - begin_index) > 1:
            self.concatenate(path, begin_index, len(path), lattice)



class Grammar:
    INHIBITED_CONNECTION = 0x7fff

    def __init__(self, bytes_, offset):
        self.POS_DEPTH = 6
        self.BOS_PARAMETER = [0, 0, 0]
        self.EOS_PARAMETER = [0, 0, 0]

        self.char_category = None

        original_offset = offset
        self.bytes = bytes_
        self.connect_table_bytes = bytes_
        self.is_copied_connect_table = False
        pos_size = self.bytes_get_short(bytes_, offset)
        offset += 2
        self.pos_list = []

        for i in range(pos_size):
            pos = []
            for j in range(self.POS_DEPTH):
                pos.append(self.buffer_to_string(offset))
                offset += 1 + 2 * len(pos[j])
            self.pos_list.append(pos)
        self.left_id_size = self.bytes_get_short(bytes_, offset)
        offset += 2
        self.right_id_size = self.bytes_get_short(bytes_, offset)
        offset += 2
        self.connect_table_offset = offset

        self.storage_size = (offset - original_offset) + 2 * self.left_id_size * self.right_id_size

    def get_storage_size(self):
        return self.storage_size

    def get_part_of_speech_size(self):
        return len(self.pos_list)

    def get_part_of_speech_string(self, pos_id):
        return self.pos_list[pos_id]

    def get_part_of_speech_id(self, pos):
        return self.pos_list.index(pos) if pos in self.pos_list else -1

    def get_connect_cost(self, left_id: int, right_id: int) -> int:
        return self.bytes_get_short(self.connect_table_bytes, self.connect_table_offset + 2 * left_id + 2 * self.left_id_size * right_id)

    def set_connect_cost(self, left_id, right_id, cost):
        # bytes_ must be ACCESS_COPY mode
        self.bytes_put_short(self.connect_table_bytes, self.connect_table_offset + 2 * left_id + 2 * self.left_id_size * right_id, cost)

    def get_bos_parameter(self):
        return self.BOS_PARAMETER

    def get_eos_parameter(self):
        return self.EOS_PARAMETER

    def get_character_category(self):
        return self.char_category

    def set_character_category(self, char_category):
        self.char_category = char_category

    def buffer_to_string(self, offset):
        self.bytes.seek(offset)
        length = self.bytes.read_byte()
        offset += 1
        string = self.bytes.read(2 * length)
        return string.decode('utf-16')

    @staticmethod
    def bytes_get_short(bytes_, offset):
        bytes_.seek(offset)
        return int.from_bytes(bytes_.read(2), 'little', signed=True)

    @staticmethod
    def bytes_put_short(bytes_, offset, data):
        bytes_.seek(offset)
        bytes_.write(data.to_bytes(2, 'little'))

    def add_pos_list(self, grammar):
        self.pos_list.extend(grammar.pos_list)
